# Copyright (c) 2008, Aldo Cortesi. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import typing
from collections import defaultdict

from libqtile import configurable
from libqtile.command.base import CommandObject
from libqtile.log_utils import logger
from libqtile.utils import has_transparency, rgb

if typing.TYPE_CHECKING:
    from libqtile.command.base import ItemT
    from libqtile.widget.base import _Widget


class Gap(CommandObject):
    """A gap placed along one of the edges of the screen

    If a gap has been defined, Qtile will avoid covering it with windows. The
    most probable reason for configuring a gap is to make space for a
    third-party bar or other static window.

    Parameters
    ==========
    size :
        The "thickness" of the gap, i.e. the height of a horizontal gap, or the
        width of a vertical gap.
    """

    def __init__(self, size):
        """ """
        # 'size' corresponds to the height of a horizontal gap, or the width
        # of a vertical gap
        self.size = size
        self.initial_size = size
        # 'length' corresponds to the width of a horizontal gap, or the height
        # of a vertical gap
        self.length = None
        self.qtile = None
        self.screen = None
        self.x = None
        self.y = None
        self.width = None
        self.height = None
        self.horizontal = None

    def _configure(self, qtile, screen, **kwargs):
        self.qtile = qtile
        self.screen = screen
        self.size = self.initial_size
        # If both horizontal and vertical gaps are present, screen corners are
        # given to the horizontal ones
        if screen.top is self:
            self.x = screen.x
            self.y = screen.y
            self.length = screen.width
            self.width = self.length
            self.height = self.initial_size
            self.horizontal = True
        elif screen.bottom is self:
            self.x = screen.x
            self.y = screen.dy + screen.dheight
            self.length = screen.width
            self.width = self.length
            self.height = self.initial_size
            self.horizontal = True
        elif screen.left is self:
            self.x = screen.x
            self.y = screen.dy
            self.length = screen.dheight
            self.width = self.initial_size
            self.height = self.length
            self.horizontal = False
        else:  # right
            self.x = screen.dx + screen.dwidth
            self.y = screen.dy
            self.length = screen.dheight
            self.width = self.initial_size
            self.height = self.length
            self.horizontal = False

    def draw(self):
        pass

    def finalize(self):
        pass

    def geometry(self):
        return (self.x, self.y, self.width, self.height)

    def _items(self, name: str) -> ItemT:
        if name == "screen" and self.screen is not None:
            return True, []
        return None

    def _select(self, name, sel):
        if name == "screen":
            return self.screen

    @property
    def position(self):
        for i in ["top", "bottom", "left", "right"]:
            if getattr(self.screen, i) is self:
                return i

    def info(self):
        return dict(position=self.position)

    def cmd_info(self):
        """
        Info for this object.
        """
        return self.info()


class Obj:
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


STRETCH = Obj("STRETCH")
CALCULATED = Obj("CALCULATED")
STATIC = Obj("STATIC")


class Bar(Gap, configurable.Configurable):
    """A bar, which can contain widgets

    Parameters
    ==========
    widgets :
        A list of widget objects.
    size :
        The "thickness" of the bar, i.e. the height of a horizontal bar, or the
        width of a vertical bar.
    """

    defaults = [
        ("background", "#000000", "Background colour."),
        ("opacity", 1, "Bar window opacity."),
        ("margin", 0, "Space around bar as int or list of ints [N E S W]."),
        ("border_color", "#000000", "Border colour as str or list of str [N E S W]"),
        ("border_width", 0, "Width of border as int of list of ints [N E S W]"),
    ]

    def __init__(self, widgets, size, **config):
        Gap.__init__(self, size)
        configurable.Configurable.__init__(self, **config)
        self.add_defaults(Bar.defaults)
        self.widgets = widgets
        self.saved_focus = None
        self.cursor_in = None
        self.window = None
        self.size_calculated = 0
        self._configured = False

        if isinstance(self.margin, int):
            self.margin = [self.margin] * 4

        if isinstance(self.border_width, int):
            self.border_width = [self.border_width] * 4

        self._initial_margin = self.margin[:]
        self.struts = [0, 0, 0, 0]
        self._add_strut = False

        self.queued_draws = 0
        self.future = None
        self._borders_drawn = False

    def _configure(self, qtile, screen, reconfigure=False):
        """
        Configure the bar. `reconfigure` is set to True when screen dimensions
        change, forcing a recalculation of the bar's dimensions.
        """
        # We only want to adjust margin sizes once unless there's a new strut or we're
        # reconfiguring the bar because the screen has changed
        if not self._configured or self._add_strut or reconfigure:
            Gap._configure(self, qtile, screen)
            self._borders_drawn = False

            if sum(self._initial_margin) or sum(self.border_width) or self._add_strut:

                try:
                    # Check if colours are valid but don't convert to rgba here
                    if isinstance(self.border_color, list) and len(self.border_color) == 4:
                        [rgb(col) for col in self.border_color]
                    else:
                        rgb(self.border_color)
                        self.border_color = [self.border_color] * 4
                except (ValueError, TypeError):
                    logger.warning(
                        "Invalid border_color specified. Borders will not be displayed."
                    )
                    self.border_width = [0, 0, 0, 0]

                # Increase the margin size for the border. The border will be drawn
                # in this space so the empty space will just be the margin.
                self.margin = [
                    m + b + s
                    for m, b, s in zip(self._initial_margin, self.border_width, self.struts)
                ]

                if self.horizontal:
                    self.x += self.margin[3] - self.border_width[3]
                    self.width -= self.margin[1] + self.margin[3]
                    self.length = self.width
                    if self.size == self.initial_size:
                        self.size += self.margin[0] + self.margin[2]
                    if self.screen.top is self:
                        self.y += self.margin[0] - self.border_width[0]
                    else:
                        self.y -= self.margin[2] + self.border_width[2]

                else:
                    self.y += self.margin[0] - self.border_width[0]
                    self.height -= self.margin[0] + self.margin[2]
                    self.length = self.height
                    self.size += self.margin[1] + self.margin[3]
                    if self.screen.left is self:
                        self.x += self.margin[3]
                    else:
                        self.x -= self.margin[1]

        width = self.width + (self.border_width[1] + self.border_width[3])
        height = self.height + (self.border_width[0] + self.border_width[2])

        if self.window:
            # We get _configure()-ed with an existing window when screens are getting
            # reconfigured but this screen is present both before and after
            self.window.place(self.x, self.y, width, height, 0, None)

        else:
            # Whereas we won't have a window if we're startup up for the first time or
            # the window has been killed by us no longer using the bar's screen

            # X11 only:
            # To preserve correct display of SysTray widget, we need a 24-bit
            # window where the user requests an opaque bar.
            if self.qtile.core.name == "x11":
                depth = (
                    32
                    if has_transparency(self.background)
                    else self.qtile.core.conn.default_screen.root_depth
                )

                self.window = self.qtile.core.create_internal(
                    self.x, self.y, width, height, depth
                )

            else:
                self.window = self.qtile.core.create_internal(self.x, self.y, width, height)

            self.window.opacity = self.opacity
            self.window.unhide()

            self.window.process_window_expose = self.process_window_expose
            self.window.process_button_click = self.process_button_click
            self.window.process_button_release = self.process_button_release
            self.window.process_pointer_enter = self.process_pointer_enter
            self.window.process_pointer_leave = self.process_pointer_leave
            self.window.process_pointer_motion = self.process_pointer_motion
            self.window.process_key_press = self.process_key_press

        # We create a new drawer even if there's already a window to ensure the
        # drawer is the right size.
        self.drawer = self.window.create_drawer(width, height)
        self.drawer.clear(self.background)

        self.crashed_widgets = []
        self.qtile.renamed_widgets = []
        if self._configured:
            for i in self.widgets:
                self._configure_widget(i)
        else:
            for idx, i in enumerate(self.widgets):
                # Create a mirror if this widget is already configured but isn't a Mirror
                # We don't do isinstance(i, Mirror) because importing Mirror (at the top)
                # would give a circular import as libqtile.widget.base imports lbqtile.bar
                if i.configured and i.__class__.__name__ != "Mirror":
                    i = i.create_mirror()
                    self.widgets[idx] = i
                success = self._configure_widget(i)
                if success:
                    qtile.register_widget(i)

        # Alert the user that we've renamed some widgets
        if self.qtile.renamed_widgets:
            logger.info(
                "The following widgets were renamed in qtile.widgets_map: %s "
                "To bind commands, rename the widget or use lazy.widget[new_name].",
                ", ".join(self.qtile.renamed_widgets),
            )
            self.qtile.renamed_widgets.clear()

        self._remove_crashed_widgets()
        self.draw()
        self._resize(self.length, self.widgets)
        self._configured = True
        self._add_strut = False

    def _configure_widget(self, widget):
        configured = True
        try:
            widget._configure(self.qtile, self)

            if self.horizontal:
                widget.offsety = self.border_width[0]
            else:
                widget.offsetx = self.border_width[3]

            widget.configured = True
        except Exception as e:
            logger.error(
                "{} widget crashed during _configure with "
                "error: {}".format(widget.__class__.__name__, repr(e))
            )
            self.crashed_widgets.append(widget)
            configured = False

        return configured

    def _remove_crashed_widgets(self):
        if self.crashed_widgets:
            from libqtile.widget.config_error import ConfigErrorWidget

        for i in self.crashed_widgets:
            index = self.widgets.index(i)
            crash = ConfigErrorWidget(widget=i)
            crash._configure(self.qtile, self)
            if self.horizontal:
                crash.offsety = self.border_width[0]
            else:
                crash.offsetx = self.border_width[3]
            self.widgets.insert(index, crash)
            self.widgets.remove(i)

    def _items(self, name: str) -> ItemT:
        if name == "screen" and self.screen is not None:
            return True, []
        elif name == "widget" and self.widgets:
            return False, [w.name for w in self.widgets]
        return None

    def _select(self, name, sel):
        if name == "screen":
            return self.screen
        elif name == "widget":
            for widget in self.widgets:
                if widget.name == sel:
                    return widget
        return None

    def finalize(self):
        self.future.cancel()
        self.drawer.finalize()
        self.window.kill()
        self.widgets.clear()

    def kill_window(self):
        """Kill the window when the bar's screen is no longer being used."""
        self.drawer.finalize()
        self.window.kill()
        self.window = None

    def _resize(self, length, widgets):
        # We want consecutive stretch widgets to split one 'block' of space between them
        stretches = []
        consecutive_stretches = defaultdict(list)
        prev_stretch = None
        for widget in widgets:
            if widget.length_type == STRETCH:
                if prev_stretch:
                    consecutive_stretches[prev_stretch].append(widget)
                else:
                    stretches.append(widget)
                    prev_stretch = widget
            else:
                prev_stretch = None

        if stretches:
            stretchspace = length - sum([i.length for i in widgets if i.length_type != STRETCH])
            stretchspace = max(stretchspace, 0)
            num_stretches = len(stretches)

            if num_stretches == 1:
                stretches[0].length = stretchspace
            else:
                block = 0
                blocks = []
                for i in widgets:
                    if i.length_type != STRETCH:
                        block += i.length
                    elif i in stretches:  # False for consecutive_stretches
                        blocks.append(block)
                        block = 0
                if block:
                    blocks.append(block)
                interval = length // num_stretches

                for idx, i in enumerate(stretches):
                    if idx == 0:
                        i.length = interval - blocks[0] - blocks[1] // 2
                    elif idx == num_stretches - 1:
                        i.length = interval - blocks[-1] - blocks[-2] // 2
                    else:
                        i.length = int(interval - blocks[idx] / 2 - blocks[idx + 1] / 2)
                    stretchspace -= i.length

                stretches[0].length += stretchspace // 2
                stretches[-1].length += stretchspace - stretchspace // 2

            for i, followers in consecutive_stretches.items():
                length = i.length // (len(followers) + 1)
                rem = i.length - length
                i.length = length
                for f in followers:
                    f.length = length
                    rem -= length
                i.length += rem

        if self.horizontal:
            offset = self.border_width[3]
            for i in widgets:
                i.offsetx = offset
                offset += i.length
        else:
            offset = self.border_width[0]
            for i in widgets:
                i.offsety = offset
                offset += i.length

    def get_widget_in_position(self, x: int, y: int) -> _Widget | None:
        if self.horizontal:
            for i in self.widgets:
                if x < i.offsetx + i.length:
                    return i
        else:
            for i in self.widgets:
                if y < i.offsety + i.length:
                    return i
        return None

    def process_button_click(self, x: int, y: int, button: int) -> None:
        # If we're clicking on a bar that's not on the current screen, focus that screen
        if self.screen is not self.qtile.current_screen:
            index = self.qtile.screens.index(self.screen)
            self.qtile.focus_screen(index, warp=False)

        widget = self.get_widget_in_position(x, y)
        if widget:
            widget.button_press(
                x - widget.offsetx,
                y - widget.offsety,
                button,
            )

    def process_button_release(self, x: int, y: int, button: int) -> None:
        widget = self.get_widget_in_position(x, y)
        if widget:
            widget.button_release(
                x - widget.offsetx,
                y - widget.offsety,
                button,
            )

    def process_pointer_enter(self, x: int, y: int) -> None:
        widget = self.get_widget_in_position(x, y)
        if widget:
            widget.mouse_enter(
                x - widget.offsetx,
                y - widget.offsety,
            )
        self.cursor_in = widget

    def process_pointer_leave(self, x: int, y: int) -> None:
        if self.cursor_in:
            self.cursor_in.mouse_leave(
                x - self.cursor_in.offsetx,
                y - self.cursor_in.offsety,
            )
            self.cursor_in = None

    def process_pointer_motion(self, x: int, y: int) -> None:
        widget = self.get_widget_in_position(x, y)
        if widget and self.cursor_in and widget is not self.cursor_in:
            self.cursor_in.mouse_leave(
                x - self.cursor_in.offsetx,
                y - self.cursor_in.offsety,
            )
            widget.mouse_enter(
                x - widget.offsetx,
                y - widget.offsety,
            )
        self.cursor_in = widget

    def process_key_press(self, keycode: int) -> None:
        if self.has_keyboard:
            self.has_keyboard.process_key_press(keycode)

    def widget_grab_keyboard(self, widget):
        """
        A widget can call this method to grab the keyboard focus
        and receive keyboard messages. When done,
        widget_ungrab_keyboard() must be called.
        """
        self.has_keyboard = widget
        self.saved_focus = self.qtile.current_window
        self.window.focus(False)

    def widget_ungrab_keyboard(self):
        """
        Removes keyboard focus from the widget.
        """
        if self.saved_focus is not None:
            self.saved_focus.focus(False)
        self.has_keyboard = None

    def process_window_expose(self):
        """
        If the window is being redrawn we need to redraw borders too.
        """
        self._borders_drawn = False
        self.draw()

    def draw(self):
        if not self.widgets:
            return  # calling self._actual_draw in this case would cause a NameError.
        if self.queued_draws == 0:
            self.future = self.qtile.call_soon(self._actual_draw)
        self.queued_draws += 1

    def _actual_draw(self):
        self.queued_draws = 0
        self._resize(self.length, self.widgets)

        # We draw the border before the widgets
        if any(self.border_width) and not self._borders_drawn:

            # The border is drawn "outside" of the bar (i.e. not in the space that the
            # widgets occupy) so we need to add the additional space
            width = self.width + self.border_width[1] + self.border_width[3]
            height = self.height + self.border_width[0] + self.border_width[2]

            # line_opts is a list of tuples where each tuple represents the borders
            # in the order N, E, S, W. The border tuple contains two pairs of
            # co-ordinates for the start and end of the border.
            line_opts = [
                ((0, self.border_width[0] * 0.5), (width, self.border_width[0] * 0.5)),
                (
                    (width - (self.border_width[1] * 0.5), self.border_width[0]),
                    (width - (self.border_width[1] * 0.5), height - self.border_width[2]),
                ),
                (
                    (0, height - self.border_width[2] + (self.border_width[2] * 0.5)),
                    (width, height - self.border_width[2] + (self.border_width[2] * 0.5)),
                ),
                (
                    (self.border_width[3] * 0.5, self.border_width[0]),
                    (self.border_width[3] * 0.5, height - self.border_width[2]),
                ),
            ]

            self.drawer.clear(self.background)

            for border_width, colour, opts in zip(
                self.border_width, self.border_color, line_opts
            ):

                if not border_width:
                    continue

                move_to, line_to = opts

                # Draw the border
                self.drawer.set_source_rgb(colour)
                self.drawer.ctx.set_line_width(border_width)
                self.drawer.ctx.move_to(*move_to)
                self.drawer.ctx.line_to(*line_to)
                self.drawer.ctx.stroke()

            self.drawer.draw(0, 0)

            # Prevent multiple redraws of borders
            self._borders_drawn = True

        for i in self.widgets:
            i.draw()
        end = i.offset + i.length  # pylint: disable=undefined-loop-variable
        # we verified that self.widgets is not empty in self.draw(), see above.
        if end < self.length:
            if self.horizontal:
                self.drawer.draw(offsetx=end, width=self.length - end)
            else:
                self.drawer.draw(offsety=end, height=self.length - end)

    def info(self):
        return dict(
            size=self.size,
            length=self.length,
            width=self.width,
            height=self.height,
            position=self.position,
            widgets=[i.info() for i in self.widgets],
            window=self.window.wid,
        )

    def is_show(self):
        return self.size != 0

    def show(self, is_show=True):
        if is_show != self.is_show():
            if is_show:
                self.size = self.size_calculated
                self._borders_drawn = False
                self.window.unhide()
            else:
                self.size_calculated = self.size
                self.size = 0
                self.window.hide()
            self.screen.group.layout_all()

    def adjust_for_strut(self, size):
        if self.size:
            self.size = self.initial_size
        for i, gap in enumerate(["top", "right", "bottom", "left"]):
            if getattr(self.screen, gap) is self:
                self.struts[i] += size

        self._add_strut = True

    def cmd_fake_button_press(self, screen, position, x, y, button=1):
        """
        Fake a mouse-button-press on the bar. Co-ordinates are relative
        to the top-left corner of the bar.

        :screen The integer screen offset
        :position One of "top", "bottom", "left", or "right"
        """
        self.process_button_click(x, y, button)


BarType = typing.Union[Bar, Gap]



# Copyright (c) 2011 Florian Mounier
# Copyright (c) 2012, 2014-2015 Tycho Andersen
# Copyright (c) 2013 Mattias Svala
# Copyright (c) 2013 Craig Barnes
# Copyright (c) 2014 ramnes
# Copyright (c) 2014 Sean Vig
# Copyright (c) 2014 Adi Sieker
# Copyright (c) 2014 Chris Wesseling
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

import libqtile.config
from libqtile import layout
from libqtile.confreader import Config
from test.layouts.layout_utils import assert_focus_path, assert_focused


class StackConfig(Config):
    auto_fullscreen = True
    groups = [
        libqtile.config.Group("a"),
        libqtile.config.Group("b"),
        libqtile.config.Group("c"),
        libqtile.config.Group("d"),
    ]
    layouts = [
        layout.Stack(num_stacks=2),
        layout.Stack(num_stacks=1),
    ]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []
    follow_mouse_focus = False


stack_config = pytest.mark.parametrize("manager", [StackConfig], indirect=True)


def _stacks(manager):
    stacks = []
    for i in manager.c.layout.info()["stacks"]:
        windows = i["clients"]
        current = i["current"]
        stacks.append(windows[current:] + windows[:current])
    return stacks


@stack_config
def test_stack_commands(manager):
    assert manager.c.layout.info()["current_stack"] == 0
    manager.test_window("one")
    assert _stacks(manager) == [["one"], []]
    assert manager.c.layout.info()["current_stack"] == 0
    manager.test_window("two")
    assert _stacks(manager) == [["one"], ["two"]]
    assert manager.c.layout.info()["current_stack"] == 1
    manager.test_window("three")
    assert _stacks(manager) == [["one"], ["three", "two"]]
    assert manager.c.layout.info()["current_stack"] == 1

    manager.c.layout.delete()
    assert _stacks(manager) == [["one", "three", "two"]]
    info = manager.c.groups()["a"]
    assert info["focus"] == "one"
    manager.c.layout.delete()
    assert len(_stacks(manager)) == 1

    manager.c.layout.add()
    assert _stacks(manager) == [["one", "three", "two"], []]

    manager.c.layout.rotate()
    assert _stacks(manager) == [[], ["one", "three", "two"]]


@stack_config
def test_stack_cmd_down(manager):
    manager.c.layout.down()


@stack_config
def test_stack_addremove(manager):
    one = manager.test_window("one")
    manager.c.layout.next()
    two = manager.test_window("two")
    three = manager.test_window("three")
    assert _stacks(manager) == [["one"], ["three", "two"]]
    assert manager.c.layout.info()["current_stack"] == 1
    manager.kill_window(three)
    assert manager.c.layout.info()["current_stack"] == 1
    manager.kill_window(two)
    assert manager.c.layout.info()["current_stack"] == 0
    manager.c.layout.next()
    two = manager.test_window("two")
    manager.c.layout.next()
    assert manager.c.layout.info()["current_stack"] == 0
    manager.kill_window(one)
    assert manager.c.layout.info()["current_stack"] == 1


@stack_config
def test_stack_rotation(manager):
    manager.c.layout.delete()
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")
    assert _stacks(manager) == [["three", "two", "one"]]
    manager.c.layout.down()
    assert _stacks(manager) == [["two", "one", "three"]]
    manager.c.layout.up()
    assert _stacks(manager) == [["three", "two", "one"]]
    manager.c.layout.down()
    manager.c.layout.down()
    assert _stacks(manager) == [["one", "three", "two"]]


@stack_config
def test_stack_nextprev(manager):
    manager.c.layout.add()
    one = manager.test_window("one")
    two = manager.test_window("two")
    three = manager.test_window("three")

    assert manager.c.groups()["a"]["focus"] == "three"
    manager.c.layout.next()
    assert manager.c.groups()["a"]["focus"] == "one"

    manager.c.layout.previous()
    assert manager.c.groups()["a"]["focus"] == "three"
    manager.c.layout.previous()
    assert manager.c.groups()["a"]["focus"] == "two"

    manager.c.layout.next()
    manager.c.layout.next()
    manager.c.layout.next()
    assert manager.c.groups()["a"]["focus"] == "two"

    manager.kill_window(three)
    manager.c.layout.next()
    assert manager.c.groups()["a"]["focus"] == "one"
    manager.c.layout.previous()
    assert manager.c.groups()["a"]["focus"] == "two"
    manager.c.layout.next()
    manager.kill_window(two)
    manager.c.layout.next()
    assert manager.c.groups()["a"]["focus"] == "one"

    manager.kill_window(one)
    manager.c.layout.next()
    assert manager.c.groups()["a"]["focus"] is None
    manager.c.layout.previous()
    assert manager.c.groups()["a"]["focus"] is None


@stack_config
def test_stack_window_removal(manager):
    manager.c.layout.next()
    manager.test_window("one")
    two = manager.test_window("two")
    manager.c.layout.down()
    manager.kill_window(two)


@stack_config
def test_stack_split(manager):
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")
    stacks = manager.c.layout.info()["stacks"]
    assert not stacks[1]["split"]
    manager.c.layout.toggle_split()
    stacks = manager.c.layout.info()["stacks"]
    assert stacks[1]["split"]


@stack_config
def test_stack_shuffle(manager):
    manager.c.next_layout()
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("three")

    stack = manager.c.layout.info()["stacks"][0]
    assert stack["clients"][stack["current"]] == "three"
    for i in range(5):
        manager.c.layout.shuffle_up()
        stack = manager.c.layout.info()["stacks"][0]
        assert stack["clients"][stack["current"]] == "three"
    for i in range(5):
        manager.c.layout.shuffle_down()
        stack = manager.c.layout.info()["stacks"][0]
        assert stack["clients"][stack["current"]] == "three"


@stack_config
def test_stack_client_to(manager):
    manager.test_window("one")
    manager.test_window("two")
    assert manager.c.layout.info()["stacks"][0]["clients"] == ["one"]
    manager.c.layout.client_to_previous()
    assert manager.c.layout.info()["stacks"][0]["clients"] == ["two", "one"]
    manager.c.layout.client_to_previous()
    assert manager.c.layout.info()["stacks"][0]["clients"] == ["one"]
    assert manager.c.layout.info()["stacks"][1]["clients"] == ["two"]
    manager.c.layout.client_to_next()
    assert manager.c.layout.info()["stacks"][0]["clients"] == ["two", "one"]


@stack_config
def test_stack_info(manager):
    manager.test_window("one")
    assert manager.c.layout.info()["stacks"]


@stack_config
def test_stack_window_focus_cycle(manager):
    # setup 3 tiled and two floating clients
    manager.test_window("one")
    manager.test_window("two")
    manager.test_window("float1")
    manager.c.window.toggle_floating()
    manager.test_window("float2")
    manager.c.window.toggle_floating()
    manager.test_window("three")

    # test preconditions, stack adds clients at pos of current
    assert manager.c.layout.info()["clients"] == ["three", "one", "two"]
    # last added window has focus
    assert_focused(manager, "three")

    # assert window focus cycle, according to order in layout
    assert_focus_path(manager, "one", "two", "float1", "float2", "three")



# Copyright (c) 2008, Aldo Cortesi. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from libqtile import ipc, sh
from libqtile.command import interface


def qshell(args) -> None:
    if args.socket is None:
        socket = ipc.find_sockfile()
    else:
        socket = args.socket
    client = ipc.Client(socket, is_json=args.is_json)
    cmd_object = interface.IPCCommandInterface(client)
    qsh = sh.QSh(cmd_object)
    if args.command is not None:
        qsh.process_line(args.command)
    else:
        qsh.loop()


def add_subcommand(subparsers, parents):
    parser = subparsers.add_parser("shell", parents=parents, help="shell-like interface to qtile")
    parser.add_argument(
        "-s",
        "--socket",
        action="store",
        type=str,
        default=None,
        help="Use specified socket to connect to qtile.",
    )
    parser.add_argument(
        "-c",
        "--command",
        action="store",
        type=str,
        default=None,
        help="Run the specified qshell command and exit.",
    )
    parser.add_argument(
        "-j",
        "--json",
        action="store_true",
        default=False,
        dest="is_json",
        help="Use json in order to communicate with qtile server.",
    )
    parser.set_defaults(func=qshell)



import json
from typing import Any
from urllib.error import URLError
from urllib.request import Request, urlopen

from libqtile.log_utils import logger
from libqtile.widget import base

try:
    import xmltodict

    def xmlparse(body):
        return xmltodict.parse(body)

except ImportError:
    # TODO: we could implement a similar parser by hand, but i'm lazy, so let's
    # punt for now
    def xmlparse(body):
        raise Exception("no xmltodict library")


class GenPollText(base.ThreadPoolText):
    """A generic text widget that polls using poll function to get the text"""

    defaults = [
        ("func", None, "Poll Function"),
    ]

    def __init__(self, **config):
        base.ThreadPoolText.__init__(self, "", **config)
        self.add_defaults(GenPollText.defaults)

    def poll(self):
        if not self.func:
            return "You need a poll function"
        return self.func()


class GenPollUrl(base.ThreadPoolText):
    """A generic text widget that polls an url and parses it using parse function"""

    defaults = [
        ("url", None, "Url"),
        ("data", None, "Post Data"),
        ("parse", None, "Parse Function"),
        ("json", True, "Is Json?"),
        ("user_agent", "Qtile", "Set the user agent"),
        ("headers", {}, "Extra Headers"),
        ("xml", False, "Is XML?"),
    ]  # type: list[tuple[str, Any, str]]

    def __init__(self, **config):
        base.ThreadPoolText.__init__(self, "", **config)
        self.add_defaults(GenPollUrl.defaults)

        self.headers["User-agent"] = self.user_agent
        if self.json:
            self.headers["Content-Type"] = "application/json"

        if self.data and not isinstance(self.data, str):
            self.data = json.dumps(self.data).encode()

    def fetch(self):
        req = Request(self.url, self.data, self.headers)
        res = urlopen(req)
        charset = res.headers.get_content_charset()

        body = res.read()
        if charset:
            body = body.decode(charset)

        if self.json:
            body = json.loads(body)

        if self.xml:
            body = xmlparse(body)
        return body

    def poll(self):
        if not self.parse or not self.url:
            return "Invalid config"

        try:
            body = self.fetch()
        except URLError:
            return "No network"

        try:
            text = self.parse(body)
        except Exception:
            logger.exception("got exception polling widget")
            text = "Can't parse"

        return text



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import sys
from importlib import reload

import pytest

from libqtile.widget import gmail_checker
from test.widgets.test_gmail_checker import FakeIMAP


@pytest.fixture
def widget(monkeypatch):
    monkeypatch.setitem(sys.modules, "imaplib", FakeIMAP("imaplib"))
    reload(gmail_checker)
    yield gmail_checker.GmailChecker


@pytest.mark.parametrize(
    "screenshot_manager",
    [
        {"username": "qtile", "password": "qtile"},
        {
            "username": "qtile",
            "password": "qtile",
            "display_fmt": "unseen[{0}]",
            "status_only_unseen": True,
        },
    ],
    indirect=True,
)
def ss_gmail_checker(screenshot_manager):
    screenshot_manager.take_screenshot()


# # This test is only required because the widget is written
# # inefficiently. display_fmt should use keys instead of indices.
# def test_gmail_checker_only_unseen(fake_qtile, monkeypatch, fake_window):
#     monkeypatch.setitem(sys.modules, "imaplib", FakeIMAP("imaplib"))
#     reload(gmail_checker)

#     gmc = gmail_checker.GmailChecker(
#         display_fmt="unseen[{0}]",
#         status_only_unseen=True,
#         username="qtile",
#         password="test"
#     )
#     fakebar = FakeBar([gmc], window=fake_window)
#     gmc._configure(fake_qtile, fakebar)
#     text = gmc.poll()
#     assert text == "unseen[2]"



# Copyright (c) 2012-2015 Tycho Andersen
# Copyright (c) 2013 xarvh
# Copyright (c) 2013 horsik
# Copyright (c) 2013-2014 roger
# Copyright (c) 2013 Tao Sauvage
# Copyright (c) 2014 ramnes
# Copyright (c) 2014 Sean Vig
# Copyright (c) 2014 Adi Sieker
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from __future__ import annotations

import contextlib
import os.path
import sys
from typing import TYPE_CHECKING

from libqtile import configurable, hook, utils
from libqtile.backend import base
from libqtile.bar import Bar
from libqtile.command.base import CommandObject

if TYPE_CHECKING:
    from typing import Callable

    from libqtile.bar import BarType
    from libqtile.command.base import ItemT
    from libqtile.group import _Group


class Key:
    """Defines a keybinding.

    Parameters
    ==========
    modifiers:
        A list of modifier specifications. Modifier specifications are one of:
        "shift", "lock", "control", "mod1", "mod2", "mod3", "mod4", "mod5".
    key:
        A key specification, e.g. "a", "Tab", "Return", "space".
    commands:
        A list of lazy command objects generated with the lazy.lazy helper.
        If multiple Call objects are specified, they are run in sequence.
    desc:
        description to be added to the key binding
    """

    def __init__(self, modifiers: list[str], key: str, *commands, desc: str = ""):
        self.modifiers = modifiers
        self.key = key
        self.commands = commands
        self.desc = desc

    def __repr__(self):
        return "<Key (%s, %s)>" % (self.modifiers, self.key)


class KeyChord:
    """Define a key chord aka vim like mode

    Parameters
    ==========
    modifiers:
        A list of modifier specifications. Modifier specifications are one of:
        "shift", "lock", "control", "mod1", "mod2", "mod3", "mod4", "mod5".
    key:
        A key specification, e.g. "a", "Tab", "Return", "space".
    submappings:
        A list of Key or KeyChord declarations to bind in this chord.
    mode:
        A string with vim like mode name. If it's set, the chord mode will
        not be left after a keystroke (except for Esc which always leaves the
        current chord/mode).
    """

    def __init__(
        self,
        modifiers: list[str],
        key: str,
        submappings: list[Key | KeyChord],
        mode: str = "",
    ):
        self.modifiers = modifiers
        self.key = key

        submappings.append(Key([], "Escape"))
        self.submappings = submappings
        self.mode = mode

    def __repr__(self):
        return "<KeyChord (%s, %s)>" % (self.modifiers, self.key)


class Mouse:
    def __init__(self, modifiers: list[str], button: str, *commands, **kwargs):
        self.modifiers = modifiers
        self.button = button
        self.commands = commands
        self.button_code = int(self.button.replace("Button", ""))
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.modmask: int = 0


class Drag(Mouse):
    """Defines binding of a mouse to some dragging action

    On each motion event command is executed with two extra parameters added x
    and y offset from previous move.
    """

    def __init__(self, *args, start=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.start = start

    def __repr__(self):
        return "<Drag (%s, %s)>" % (self.modifiers, self.button)


class Click(Mouse):
    """Defines binding of a mouse click"""

    def __repr__(self):
        return "<Click (%s, %s)>" % (self.modifiers, self.button)


class EzConfig:
    """
    Helper class for defining key and button bindings in an emacs-like format.
    Inspired by Xmonad's XMonad.Util.EZConfig.
    """

    modifier_keys = {
        "M": "mod4",
        "A": "mod1",
        "S": "shift",
        "C": "control",
    }

    def parse(self, spec):
        """
        Splits an emacs keydef into modifiers and keys. For example:
          "M-S-a"     -> ['mod4', 'shift'], 'a'
          "A-<minus>" -> ['mod1'], 'minus'
          "C-<Tab>"   -> ['control'], 'Tab'
        """
        mods = []
        keys = []

        for key in spec.split("-"):
            if not key:
                break
            if key in self.modifier_keys:
                if keys:
                    msg = "Modifiers must always come before key/btn: %s"
                    raise utils.QtileError(msg % spec)
                mods.append(self.modifier_keys[key])
                continue
            if len(key) == 1:
                keys.append(key)
                continue
            if len(key) > 3 and key[0] == "<" and key[-1] == ">":
                keys.append(key[1:-1])
                continue

        if not keys:
            msg = "Invalid key/btn specifier: %s"
            raise utils.QtileError(msg % spec)

        if len(keys) > 1:
            msg = "Key chains are not supported: %s" % spec
            raise utils.QtileError(msg)

        return mods, keys[0]


class EzKey(EzConfig, Key):
    def __init__(self, keydef, *commands, **kwargs):
        modkeys, key = self.parse(keydef)
        super().__init__(modkeys, key, *commands, **kwargs)


class EzClick(EzConfig, Click):
    def __init__(self, btndef, *commands, **kwargs):
        modkeys, button = self.parse(btndef)
        button = "Button%s" % button
        super().__init__(modkeys, button, *commands, **kwargs)


class EzDrag(EzConfig, Drag):
    def __init__(self, btndef, *commands, **kwargs):
        modkeys, button = self.parse(btndef)
        button = "Button%s" % button
        super().__init__(modkeys, button, *commands, **kwargs)


class ScreenRect:
    def __init__(self, x, y, width, height):
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def __repr__(self):
        return "<%s %d,%d %d,%d>" % (
            self.__class__.__name__,
            self.x,
            self.y,
            self.width,
            self.height,
        )

    def hsplit(self, columnwidth):
        assert columnwidth > 0
        assert columnwidth < self.width
        return (
            self.__class__(self.x, self.y, columnwidth, self.height),
            self.__class__(self.x + columnwidth, self.y, self.width - columnwidth, self.height),
        )

    def vsplit(self, rowheight):
        assert rowheight > 0
        assert rowheight < self.height
        return (
            self.__class__(self.x, self.y, self.width, rowheight),
            self.__class__(self.x, self.y + rowheight, self.width, self.height - rowheight),
        )


class Screen(CommandObject):
    """A physical screen, and its associated paraphernalia.

    Define a screen with a given set of Bars of a specific geometry.  Note that
    bar.Bar objects can only be placed at the top or the bottom of the screen
    (bar.Gap objects can be placed anywhere).  Also, ``x``, ``y``, ``width``,
    and ``height`` aren't specified usually unless you are using 'fake
    screens'.

    The ``wallpaper`` parameter, if given, should be a path to an image file. How this
    image is painted to the screen is specified by the ``wallpaper_mode`` parameter. By
    default, the image will be placed at the screens origin and retain its own
    dimensions. If the mode is 'fill', the image will be centred on the screen and
    resized to fill it. If the mode is 'stretch', the image is stretched to fit all of
    it into the screen.
    """

    group: _Group
    previous_group: _Group
    index: int

    def __init__(
        self,
        top: BarType | None = None,
        bottom: BarType | None = None,
        left: BarType | None = None,
        right: BarType | None = None,
        wallpaper: str | None = None,
        wallpaper_mode: str | None = None,
        x: int | None = None,
        y: int | None = None,
        width: int | None = None,
        height: int | None = None,
    ):

        self.top = top
        self.bottom = bottom
        self.left = left
        self.right = right
        self.wallpaper = wallpaper
        self.wallpaper_mode = wallpaper_mode
        self.qtile = None
        # x position of upper left corner can be > 0
        # if one screen is "right" of the other
        self.x = x if x is not None else 0
        self.y = y if y is not None else 0
        self.width = width if width is not None else 0
        self.height = height if height is not None else 0

    def _configure(self, qtile, index, x, y, width, height, group, reconfigure_gaps=False):
        self.qtile = qtile
        self.index = index
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.set_group(group)
        for i in self.gaps:
            i._configure(qtile, self, reconfigure=reconfigure_gaps)
        if self.wallpaper:
            self.wallpaper = os.path.expanduser(self.wallpaper)
            self.paint(self.wallpaper, self.wallpaper_mode)

    def paint(self, path, mode=None):
        self.qtile.paint_screen(self, path, mode)

    @property
    def gaps(self):
        return (i for i in [self.top, self.bottom, self.left, self.right] if i)

    @property
    def dx(self):
        return self.x + self.left.size if self.left else self.x

    @property
    def dy(self):
        return self.y + self.top.size if self.top else self.y

    @property
    def dwidth(self):
        val = self.width
        if self.left:
            val -= self.left.size
        if self.right:
            val -= self.right.size
        return val

    @property
    def dheight(self):
        val = self.height
        if self.top:
            val -= self.top.size
        if self.bottom:
            val -= self.bottom.size
        return val

    def get_rect(self):
        return ScreenRect(self.dx, self.dy, self.dwidth, self.dheight)

    def set_group(self, new_group, save_prev=True, warp=True):
        """Put group on this screen"""
        if new_group is None:
            return

        if new_group.screen == self:
            return

        if save_prev and hasattr(self, "group"):
            self.previous_group = self.group

        if new_group.screen:
            # g1 <-> s1 (self)
            # g2 (new_group) <-> s2 to
            # g1 <-> s2
            # g2 <-> s1
            g1 = self.group
            s1 = self
            g2 = new_group
            s2 = new_group.screen

            s2.group = g1
            g1.set_screen(s2, warp)
            s1.group = g2
            g2.set_screen(s1, warp)
        else:
            if hasattr(self, "group"):
                old_group = self.group
                ctx = self.qtile.core.masked()
            else:
                old_group = None
                ctx = contextlib.nullcontext()
            self.group = new_group
            with ctx:
                # display clients of the new group and then hide from old group
                # to remove the screen flickering
                new_group.set_screen(self, warp)

                if old_group is not None:
                    old_group.set_screen(None, warp)

        hook.fire("setgroup")
        hook.fire("focus_change")
        hook.fire("layout_change", self.group.layouts[self.group.current_layout], self.group)

    def toggle_group(self, group=None, warp=True):
        """Switch to the selected group or to the previously active one"""
        if group in (self.group, None) and hasattr(self, "previous_group"):
            group = self.previous_group
        self.set_group(group, warp=warp)

    def _items(self, name: str) -> ItemT:
        if name == "layout" and self.group is not None:
            return True, list(range(len(self.group.layouts)))
        elif name == "window" and self.group is not None:
            return True, [i.wid for i in self.group.windows]
        elif name == "bar":
            return False, [x.position for x in self.gaps]
        elif name == "widget":
            return False, [w.name for g in self.gaps for w in g.widgets if isinstance(g, Bar)]
        elif name == "group":
            return True, [self.group.name]
        return None

    def _select(self, name, sel):
        if name == "layout":
            if sel is None:
                return self.group.layout
            else:
                return utils.lget(self.group.layouts, sel)
        elif name == "window":
            if sel is None:
                return self.group.current_window
            else:
                for i in self.group.windows:
                    if i.wid == sel:
                        return i
        elif name == "bar":
            return getattr(self, sel)
        elif name == "widget":
            for gap in self.gaps:
                if not isinstance(gap, Bar):
                    continue
                for widget in gap.widgets:
                    if widget.name == sel:
                        return widget
        elif name == "group":
            if sel is None:
                return self.group
            else:
                return self.group if sel == self.group.name else None

    def resize(self, x=None, y=None, w=None, h=None):
        if x is None:
            x = self.x
        if y is None:
            y = self.y
        if w is None:
            w = self.width
        if h is None:
            h = self.height
        self._configure(self.qtile, self.index, x, y, w, h, self.group)
        for bar in [self.top, self.bottom, self.left, self.right]:
            if bar:
                bar.draw()
        self.qtile.call_soon(self.group.layout_all)

    def cmd_info(self):
        """Returns a dictionary of info for this screen."""
        return dict(index=self.index, width=self.width, height=self.height, x=self.x, y=self.y)

    def cmd_resize(self, x=None, y=None, w=None, h=None):
        """Resize the screen"""
        self.resize(x, y, w, h)

    def cmd_next_group(self, skip_empty=False, skip_managed=False):
        """Switch to the next group"""
        n = self.group.get_next_group(skip_empty, skip_managed)
        self.set_group(n)
        return n.name

    def cmd_prev_group(self, skip_empty=False, skip_managed=False, warp=True):
        """Switch to the previous group"""
        n = self.group.get_previous_group(skip_empty, skip_managed)
        self.set_group(n, warp=warp)
        return n.name

    def cmd_toggle_group(self, group_name=None, warp=True):
        """Switch to the selected group or to the previously active one"""
        group = self.qtile.groups_map.get(group_name)
        self.toggle_group(group, warp=warp)

    def cmd_set_wallpaper(self, path, mode=None):
        """Set the wallpaper to the given file."""
        self.paint(path, mode)


class Group:
    """Represents a "dynamic" group

    These groups can spawn apps, only allow certain Matched windows to be on
    them, hide when they're not in use, etc.
    Groups are identified by their name.

    Parameters
    ==========
    name: string
        the name of this group
    matches: default ``None``
        list of ``Match`` objects whose  windows will be assigned to this group
    exclusive: boolean
        when other apps are started in this group, should we allow them here or not?
    spawn: string or list of strings
        this will be ``exec()`` d when the group is created, you can pass
        either a program name or a list of programs to ``exec()``
    layout: string
        the name of default layout for this group (e.g. 'max' or 'stack').
        This is the name specified for a particular layout in config.py
        or if not defined it defaults in general the class name in all lower case.
    layouts: list
        the group layouts list overriding global layouts.
        Use this to define a separate list of layouts for this particular group.
    persist: boolean
        should this group stay alive with no member windows?
    init: boolean
        is this group alive when qtile starts?
    position  int
        group position
    label: string
        the display name of the group.
        Use this to define a display name other than name of the group.
        If set to None, the display name is set to the name.
    """

    def __init__(
        self,
        name: str,
        matches: list[Match] | None = None,
        exclusive=False,
        spawn: str | list[str] | None = None,
        layout: str | None = None,
        layouts: list | None = None,
        persist=True,
        init=True,
        layout_opts=None,
        screen_affinity=None,
        position=sys.maxsize,
        label: str | None = None,
    ):
        self.name = name
        self.label = label
        self.exclusive = exclusive
        self.spawn = spawn
        self.layout = layout
        self.layouts = layouts or []
        self.persist = persist
        self.init = init
        self.matches = matches or []
        self.layout_opts = layout_opts or {}

        self.screen_affinity = screen_affinity
        self.position = position

    def __repr__(self):
        attrs = utils.describe_attributes(
            self,
            [
                "exclusive",
                "spawn",
                "layout",
                "layouts",
                "persist",
                "init",
                "matches",
                "layout_opts",
                "screen_affinity",
            ],
        )
        return "<config.Group %r (%s)>" % (self.name, attrs)


class ScratchPad(Group):
    """Represents a "ScratchPad" group

    ScratchPad adds a (by default) invisible group to qtile.
    That group is used as a place for currently not visible windows spawned by a
    ``DropDown`` configuration.

    Parameters
    ==========
    name: string
        the name of this group
    dropdowns: default ``None``
        list of DropDown objects
    position: int
        group position
    label: string
        The display name of the ScratchPad group. Defaults to the empty string
        such that the group is hidden in ``GroupList`` widget.
    single : Boolean
        Only one of the window among the specified dropdowns will be
        visible at a time.
    """

    def __init__(self, name, dropdowns=None, position=sys.maxsize, label="", single=False):
        Group.__init__(
            self,
            name,
            layout="floating",
            layouts=["floating"],
            init=False,
            position=position,
            label=label,
        )
        self.dropdowns = dropdowns if dropdowns is not None else []
        self.single = single

    def __repr__(self):
        return "<config.ScratchPad %r (%s)>" % (
            self.name,
            ", ".join(dd.name for dd in self.dropdowns),
        )


class Match:
    """Match for dynamic groups or auto-floating windows.

    It can match by title, wm_class, role, wm_type, wm_instance_class or
    net_wm_pid.

    ``Match`` supports both regular expression objects (i.e. the result of
    ``re.compile()``) or strings (match as an "include"-match). If a window
    matches all specified values, it is considered a match.

    Parameters
    ==========
    title:
        matches against the WM_NAME atom (X11) or title (Wayland)
    wm_class:
        matches against the second string in WM_CLASS atom (X11) or app ID (Wayland)
    role:
        matches against the WM_ROLE atom (X11 only)
    wm_type:
        matches against the WM_TYPE atom (X11 only)
    wm_instance_class:
        matches against the first string in WM_CLASS atom (X11) or app ID (Wayland)
    net_wm_pid:
        matches against the _NET_WM_PID atom (X11) or PID (Wayland) -
        (only int allowed for this rule)
    func:
        delegate the match to the given function, which receives the tested
        client as argument and must return True if it matches, False otherwise
    """

    def __init__(
        self,
        title=None,
        wm_class=None,
        role=None,
        wm_type=None,
        wm_instance_class=None,
        net_wm_pid=None,
        func: Callable[[base.WindowType], bool] | None = None,
        wid=None,
    ):
        self._rules = {}

        if title is not None:
            self._rules["title"] = title
        if wm_class is not None:
            self._rules["wm_class"] = wm_class
        if wm_instance_class is not None:
            self._rules["wm_instance_class"] = wm_instance_class
        if wid is not None:
            self._rules["wid"] = wid
        if net_wm_pid is not None:
            try:
                self._rules["net_wm_pid"] = int(net_wm_pid)
            except ValueError:
                error = 'Invalid rule for net_wm_pid: "%s" only int allowed' % str(net_wm_pid)
                raise utils.QtileError(error)
        if func is not None:
            self._rules["func"] = func

        if role is not None:
            self._rules["role"] = role
        if wm_type is not None:
            self._rules["wm_type"] = wm_type

    @staticmethod
    def _get_property_predicate(name, value):
        if name == "net_wm_pid" or name == "wid":
            return lambda other: other == value
        elif name == "wm_class":

            def predicate(other):
                # match as an "include"-match on any of the received classes
                match = getattr(other, "match", lambda v: v in other)
                return value and any(match(v) for v in value)

            return predicate
        else:

            def predicate(other):
                # match as an "include"-match
                match = getattr(other, "match", lambda v: v in other)
                return match(value)

            return predicate

    def compare(self, client):
        for property_name, rule_value in self._rules.items():
            if property_name == "title":
                value = client.name
            elif "class" in property_name:
                wm_class = client.get_wm_class()
                if not wm_class:
                    return False
                if property_name == "wm_instance_class":
                    value = wm_class[0]
                else:
                    value = wm_class
            elif property_name == "role":
                value = client.get_wm_role()
            elif property_name == "func":
                return rule_value(client)
            elif property_name == "net_wm_pid":
                value = client.get_pid()
            elif property_name == "wid":
                value = client.wid
            else:
                value = client.get_wm_type()

            # Some of the window.get_...() functions can return None
            if value is None:
                return False

            match = self._get_property_predicate(property_name, value)
            if not match(rule_value):
                return False

        if not self._rules:
            return False
        return True

    def map(self, callback, clients):
        """Apply callback to each client that matches this Match"""
        for c in clients:
            if self.compare(c):
                callback(c)

    def __repr__(self):
        return "<Match %s>" % self._rules


class Rule:
    """How to act on a match

    A Rule contains a list of Match objects, and a specification about what to
    do when any of them is matched.

    Parameters
    ==========
    match :
        ``Match`` object or a list of such associated with this ``Rule``
    float :
        auto float this window?
    intrusive :
        override the group's exclusive setting?
    break_on_match :
        Should we stop applying rules if this rule is matched?
    """

    def __init__(self, match, group=None, float=False, intrusive=False, break_on_match=True):
        if isinstance(match, Match):
            self.matchlist = [match]
        else:
            self.matchlist = match
        self.group = group
        self.float = float
        self.intrusive = intrusive
        self.break_on_match = break_on_match

    def matches(self, w):
        return any(w.match(m) for m in self.matchlist)

    def __repr__(self):
        actions = utils.describe_attributes(
            self, ["group", "float", "intrusive", "break_on_match"]
        )
        return "<Rule match=%r actions=(%s)>" % (self.matchlist, actions)


class DropDown(configurable.Configurable):
    """
    Configure a specified command and its associated window for the ScratchPad.
    That window can be shown and hidden using a configurable keystroke
    or any other scripted trigger.
    """

    defaults = (
        (
            "x",
            0.1,
            "X position of window as fraction of current screen width. "
            "0 is the left most position.",
        ),
        (
            "y",
            0.0,
            "Y position of window as fraction of current screen height. "
            "0 is the top most position. To show the window at bottom, "
            "you have to configure a value < 1 and an appropriate height.",
        ),
        ("width", 0.8, "Width of window as fraction of current screen width"),
        ("height", 0.35, "Height of window as fraction of current screen."),
        ("opacity", 0.9, "Opacity of window as fraction. Zero is opaque."),
        (
            "on_focus_lost_hide",
            True,
            "Shall the window be hidden if focus is lost? If so, the DropDown "
            "is hidden if window focus or the group is changed.",
        ),
        (
            "warp_pointer",
            True,
            "Shall pointer warp to center of window on activation? "
            "This has only effect if any of the on_focus_lost_xxx "
            "configurations is True",
        ),
        (
            "match",
            None,
            "Use a ``config.Match`` to identify the spawned window and move it to the "
            "scratchpad, instead of relying on the window's PID. This works around "
            "some programs that may not be caught by the window's PID if it does "
            "not match the PID of the spawned process.",
        ),
    )

    def __init__(self, name, cmd, **config):
        """
        Initialize DropDown window wrapper.
        Define a command to spawn a process for the first time the DropDown
        is shown.

        Parameters
        ==========
        name: string
            The name of the DropDown configuration.
        cmd: string
            Command to spawn a process.
        match : Match
            A match object to identify the window instead of the pid.
        """
        configurable.Configurable.__init__(self, **config)
        self.name = name
        self.command = cmd
        self.add_defaults(self.defaults)

    def info(self):
        return dict(
            name=self.name,
            command=self.command,
            x=self.x,
            y=self.y,
            width=self.width,
            height=self.height,
            opacity=self.opacity,
            on_focus_lost_hide=self.on_focus_lost_hide,
            warp_pointer=self.warp_pointer,
        )



# Copyright (c) 2021 Graeme Holliday
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from dbus_next.aio import MessageBus
from dbus_next.constants import BusType

from libqtile.widget import base

BLUEZ = "org.bluez"
BLUEZ_PATH = "/org/bluez/hci0"
BLUEZ_ADAPTER = "org.bluez.Adapter1"
BLUEZ_DEVICE = "org.bluez.Device1"
BLUEZ_PROPERTIES = "org.freedesktop.DBus.Properties"


class Bluetooth(base._TextBox):
    """
    Displays bluetooth status or connected device.

    Uses dbus to communicate with the system bus.

    Widget requirements: dbus-next_.

    .. _dbus-next: https://pypi.org/project/dbus-next/
    """

    defaults = [
        (
            "hci",
            "/dev_XX_XX_XX_XX_XX_XX",
            "hci0 device path, can be found with d-feet or similar dbus explorer.",
        )
    ]

    def __init__(self, **config):
        base._TextBox.__init__(self, "", **config)
        self.add_defaults(Bluetooth.defaults)

    async def _config_async(self):
        # set initial values
        self.powered = await self._init_adapter()
        self.connected, self.device = await self._init_device()

        self.update_text()

    async def _init_adapter(self):
        # set up interface to adapter properties using high-level api
        bus = await MessageBus(bus_type=BusType.SYSTEM).connect()
        introspect = await bus.introspect(BLUEZ, BLUEZ_PATH)
        obj = bus.get_proxy_object(BLUEZ, BLUEZ_PATH, introspect)
        iface = obj.get_interface(BLUEZ_ADAPTER)
        props = obj.get_interface(BLUEZ_PROPERTIES)

        powered = await iface.get_powered()
        # subscribe receiver to property changed
        props.on_properties_changed(self._signal_received)
        return powered

    async def _init_device(self):
        # set up interface to device properties using high-level api
        bus = await MessageBus(bus_type=BusType.SYSTEM).connect()
        introspect = await bus.introspect(BLUEZ, BLUEZ_PATH + self.hci)
        obj = bus.get_proxy_object(BLUEZ, BLUEZ_PATH + self.hci, introspect)
        iface = obj.get_interface(BLUEZ_DEVICE)
        props = obj.get_interface(BLUEZ_PROPERTIES)

        connected = await iface.get_connected()
        name = await iface.get_name()
        # subscribe receiver to property changed
        props.on_properties_changed(self._signal_received)
        return connected, name

    def _signal_received(self, interface_name, changed_properties, _invalidated_properties):
        powered = changed_properties.get("Powered", None)
        if powered is not None:
            self.powered = powered.value
            self.update_text()

        connected = changed_properties.get("Connected", None)
        if connected is not None:
            self.connected = connected.value
            self.update_text()

        device = changed_properties.get("Name", None)
        if device is not None:
            self.device = device.value
            self.update_text()

    def update_text(self):
        text = ""
        if not self.powered:
            text = "off"
        else:
            if not self.connected:
                text = "on"
            else:
                text = self.device
        self.update(text)



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import json
import os
import shutil
from functools import partial
from pathlib import Path

import cairocffi
import pytest

from libqtile.bar import Bar
from libqtile.config import Group, Screen


@pytest.fixture(scope="session")
def target():
    folder = Path(__file__).parent / "screenshots"
    docs_folder = (
        Path(__file__).parent
        / ".."
        / ".."
        / ".."
        / "docs"
        / "_static"
        / "screenshots"
        / "widgets"
    )
    log = os.path.join(docs_folder, "shots.json")
    if folder.is_dir():
        shutil.rmtree(folder)
    folder.mkdir()
    key = {}

    def get_file_name(w_name, config):
        nonlocal key

        # Convert config into a string of key=value
        entry = ", ".join(f"{k}={repr(v)}" for k, v in config.items())

        # Check if widget is in the key dict
        if w_name not in key:
            key[w_name] = {}

        # Increment the index number
        indexes = [int(x) for x in key[w_name]]
        index = max(indexes) + 1 if indexes else 1

        # Record the config
        key[w_name][index] = entry

        # Define the target folder and check it exists
        shots_dir = os.path.join(folder, w_name)
        if not os.path.isdir(shots_dir):
            os.mkdir(shots_dir)

        # Returnt the path for the screenshot
        return os.path.join(shots_dir, f"{index}.png")

    yield get_file_name

    # We copy the screenshots from the test folder to the docs folder at the end
    # This prevents pytest deleting the files itself

    # Remove old screenshots
    if os.path.isdir(docs_folder):
        shutil.rmtree(docs_folder)

    # Copy to the docs folder
    shutil.copytree(folder, docs_folder)
    with open(log, "w") as f:
        json.dump(key, f)

    # Clear up the tests folder
    shutil.rmtree(folder)


@pytest.fixture
def screenshot_manager(widget, request, manager_nospawn, minimal_conf_noscreen, target):
    """
    Create a manager instance for the screenshots. Individual "tests" should only call
    `screenshot_manager.take_screenshot()` but the destination path is also available in
    `screenshot_manager.target`.

    Widgets should create their own `widget` fixture in the relevant file (applying
    monkeypatching etc as necessary).

    Configs can then be passed by parametrizing "screenshot_manager".
    """
    # Partials are used to hide some aspects of the config from being displayed in the
    # docs. We need to split these out into their constituent parts.
    if type(widget) is partial:
        widget_class = widget.func
        widget_config = widget.keywords
    else:
        widget_class = widget
        widget_config = {}

    class ScreenshotWidget(widget_class):
        def __init__(self, *args, **kwargs):
            widget_class.__init__(self, *args, **kwargs)
            # We need the widget's name to be the name of the inherited class
            self.name = widget_class.__name__.lower()

        def cmd_take_screenshot(self, target):
            if not self.configured:
                return

            # Screenshots only run on X11
            source = self.drawer._xcb_surface

            dest = cairocffi.ImageSurface(cairocffi.FORMAT_ARGB32, self.width, self.height)
            with cairocffi.Context(dest) as ctx:
                ctx.set_source_surface(source)
                ctx.paint()

            dest.write_to_png(target)

    class ScreenshotBar(Bar):
        def cmd_take_screenshot(self, target, x=0, y=0, width=None, height=None):
            """Takes a screenshot of the bar. The area can be selected."""
            if not self._configured:
                return

            if width is None:
                width = self.drawer.width

            if height is None:
                height = self.drawer.height

            # Screenshots only run on X11
            source = "_xcb_surface"

            # Widgets aren't drawn to the bar's drawer so we first need to render them all to a single surface
            bar_copy = cairocffi.ImageSurface(
                cairocffi.FORMAT_ARGB32, self.drawer.width, self.drawer.height
            )
            with cairocffi.Context(bar_copy) as ctx:
                ctx.set_source_surface(getattr(self.drawer, source))
                ctx.paint()

                for i in self.widgets:
                    ctx.set_source_surface(getattr(i.drawer, source), i.offsetx, i.offsety)
                    ctx.paint()

            # Then we copy the desired area to our destination surface
            dest = cairocffi.ImageSurface(cairocffi.FORMAT_ARGB32, width, height)
            with cairocffi.Context(dest) as ctx:
                ctx.set_source_surface(bar_copy, x=x, y=y)
                ctx.paint()

            dest.write_to_png(target)

    # Get the widget and config
    config = getattr(request, "param", dict())
    wdgt = ScreenshotWidget(**{**widget_config, **config})
    name = wdgt.name

    # Create a function to generate filename
    def filename():
        return target(name, config)

    # Add the widget to our config
    minimal_conf_noscreen.groups = [Group(i) for i in "123456789"]
    minimal_conf_noscreen.fake_screens = [
        Screen(top=ScreenshotBar([wdgt], 32), x=0, y=0, width=300, height=300),
        Screen(top=ScreenshotBar([], 32), x=0, y=300, width=300, height=300),
    ]

    manager_nospawn.start(minimal_conf_noscreen)

    # Add some convenience attributes for taking screenshots
    manager_nospawn.target = filename
    ss_widget = manager_nospawn.c.widget[name]
    manager_nospawn.take_screenshot = lambda f=filename: ss_widget.take_screenshot(f())

    yield manager_nospawn



# -*- coding: utf-8 -*-
# Copyright (c) 2011 Florian Mounier
# Copyright (c) 2011 Mounier Florian
# Copyright (c) 2012 roger
# Copyright (c) 2012-2014 Tycho Andersen
# Copyright (c) 2012-2013 Craig Barnes
# Copyright (c) 2013 Tao Sauvage
# Copyright (c) 2014 Sean Vig
# Copyright (c) 2014 Adi Sieker
# Copyright (c) 2020 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import asyncio
from os import path

from libqtile import bar, pangocffi, utils
from libqtile.log_utils import logger
from libqtile.notify import ClosedReason, notifier
from libqtile.widget import base


class Notify(base._TextBox):
    """
    A notify widget

    This widget can handle actions provided by notification clients. However, only the
    default action is supported, so if a client provides multiple actions then only the
    default (first) action can be invoked. Some programs will provide their own
    notification windows if the notification server does not support actions, so if you
    want your notifications to handle more than one action then specify ``False`` for
    the ``action`` option to disable all action handling. Unfortunately we cannot
    specify the capability for exactly one action.
    """

    defaults = [
        ("foreground_urgent", "ff0000", "Foreground urgent priority colour"),
        ("foreground_low", "dddddd", "Foreground low priority  colour"),
        ("default_timeout", None, "Default timeout (seconds) for notifications"),
        ("audiofile", None, "Audiofile played during notifications"),
        ("action", True, "Enable handling of default action upon right click"),
        (
            "parse_text",
            None,
            "Function to parse and modify notifications. "
            "e.g. function in config that removes line returns:"
            "def my_func(text)"
            "   return text.replace('\n', '')"
            "then set option parse_text=my_func",
        ),
    ]
    capabilities = {"body", "actions"}

    def __init__(self, width=bar.CALCULATED, **config):
        base._TextBox.__init__(self, "", width, **config)
        self.add_defaults(Notify.defaults)
        self.current_id = 0

        default_callbacks = {
            "Button1": self.clear,
            "Button4": self.prev,
            "Button5": self.next,
        }
        if self.action:
            default_callbacks["Button3"] = self.invoke
        else:
            self.capabilities = Notify.capabilities.difference({"actions"})
        self.add_callbacks(default_callbacks)

    def _configure(self, qtile, bar):
        base._TextBox._configure(self, qtile, bar)
        self.layout = self.drawer.textlayout(
            self.text, self.foreground, self.font, self.fontsize, self.fontshadow, markup=True
        )

    async def _config_async(self):
        await notifier.register(self.update, self.capabilities, on_close=self.on_close)

    def set_notif_text(self, notif):
        self.text = pangocffi.markup_escape_text(notif.summary)
        urgency = getattr(notif.hints.get("urgency"), "value", 1)
        if urgency != 1:
            self.text = '<span color="%s">%s</span>' % (
                utils.hex(self.foreground_urgent if urgency == 2 else self.foreground_low),
                self.text,
            )
        if notif.body:
            self.text = '<span weight="bold">%s</span> - %s' % (
                self.text,
                pangocffi.markup_escape_text(notif.body),
            )
        if callable(self.parse_text):
            try:
                self.text = self.parse_text(self.text)
            except:  # noqa: E722
                logger.exception("parse_text function failed:")
        if self.audiofile and path.exists(self.audiofile):
            self.qtile.cmd_spawn("aplay -q '%s'" % self.audiofile)

    def update(self, notif):
        self.qtile.call_soon_threadsafe(self.real_update, notif)

    def real_update(self, notif):
        self.set_notif_text(notif)
        self.current_id = notif.id - 1
        if notif.timeout and notif.timeout > 0:
            self.timeout_add(
                notif.timeout / 1000, self.clear, method_args=(ClosedReason.expired,)
            )
        elif self.default_timeout:
            self.timeout_add(
                self.default_timeout, self.clear, method_args=(ClosedReason.expired,)
            )
        self.bar.draw()
        return True

    def display(self):
        self.set_notif_text(notifier.notifications[self.current_id])
        self.bar.draw()

    def clear(self, reason=ClosedReason.dismissed):
        notifier._service.NotificationClosed(notifier.notifications[self.current_id].id, reason)
        self.text = ""
        self.current_id = len(notifier.notifications) - 1
        self.bar.draw()

    def on_close(self, nid):
        if self.current_id < len(notifier.notifications):
            notif = notifier.notifications[self.current_id]
            if notif.id == nid:
                self.clear(ClosedReason.method)

    def prev(self):
        if self.current_id > 0:
            self.current_id -= 1
        self.display()

    def next(self):
        if self.current_id < len(notifier.notifications) - 1:
            self.current_id += 1
            self.display()

    def invoke(self):
        if self.current_id < len(notifier.notifications):
            notif = notifier.notifications[self.current_id]
            if notif.actions:
                notifier._service.ActionInvoked(notif.id, notif.actions[0])
            self.clear()

    def cmd_display(self):
        """Display the notifcication"""
        self.display()

    def cmd_clear(self):
        """Clear the notification"""
        self.clear()

    def cmd_toggle(self):
        """Toggle showing/clearing the notification"""
        if self.text == "":
            self.display()
        else:
            self.clear()

    def cmd_prev(self):
        """Show previous notification"""
        self.prev()

    def cmd_next(self):
        """Show next notification"""
        self.next()

    def cmd_invoke(self):
        """Invoke the notification's default action"""
        if self.action:
            self.invoke()

    def finalize(self):
        asyncio.create_task(self._finalize())

    async def _finalize(self):
        task = notifier.unregister(self.update)

        # If the notifier has no more callbacks then it needs to be stopped.
        # The returned task will handle the release of the service name from
        # dbus. We await it here to make sure it's finished before we
        # complete the finalisation of this widget.
        if task:
            await task
        base._TextBox.finalize(self)



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# This widget is based on GenPollUrl which has a separate.
# We just need to test parsing here.

from libqtile import widget

RESPONSE = {"data": {"base": "BTC", "currency": "GBP", "amount": "29625.02"}}


def test_set_defaults():
    crypto = widget.CryptoTicker(currency="", symbol="")
    assert crypto.currency == "USD"
    assert crypto.symbol == "$"


def test_parse():
    crypto = widget.CryptoTicker(currency="GBP", symbol="£", crypto="BTC")
    assert crypto.url == "https://api.coinbase.com/v2/prices/BTC-GBP/spot"
    assert crypto.parse(RESPONSE) == "BTC: £29625.02"



# Copyright (c) 2011 Florian Mounier
# Copyright (c) 2012, 2014-2015 Tycho Andersen
# Copyright (c) 2013 Mattias Svala
# Copyright (c) 2013 Craig Barnes
# Copyright (c) 2014 ramnes
# Copyright (c) 2014 Sean Vig
# Copyright (c) 2014 Adi Sieker
# Copyright (c) 2014 Chris Wesseling
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

import libqtile.config
from libqtile import layout
from libqtile.config import Match
from libqtile.confreader import Config
from test.layouts.layout_utils import assert_dimensions, assert_focus_path, assert_focused


class SliceConfig(Config):
    auto_fullscreen = True
    groups = [
        libqtile.config.Group("a"),
    ]

    layouts = [
        layout.Slice(
            side="left",
            width=200,
            match=Match(title="slice"),
            fallback=layout.Stack(num_stacks=1, border_width=0),
        ),
        layout.Slice(
            side="right",
            width=200,
            match=Match(title="slice"),
            fallback=layout.Stack(num_stacks=1, border_width=0),
        ),
        layout.Slice(
            side="top",
            width=200,
            match=Match(title="slice"),
            fallback=layout.Stack(num_stacks=1, border_width=0),
        ),
        layout.Slice(
            side="bottom",
            width=200,
            match=Match(title="slice"),
            fallback=layout.Stack(num_stacks=1, border_width=0),
        ),
    ]
    floating_layout = libqtile.resources.default_config.floating_layout
    keys = []
    mouse = []
    screens = []
    follow_mouse_focus = False


slice_config = pytest.mark.parametrize("manager", [SliceConfig], indirect=True)


@slice_config
def test_no_slice(manager):
    manager.test_window("one")
    assert_dimensions(manager, 200, 0, 600, 600)
    manager.test_window("two")
    assert_dimensions(manager, 200, 0, 600, 600)


@slice_config
def test_slice_first(manager):
    manager.test_window("slice")
    assert_dimensions(manager, 0, 0, 200, 600)
    manager.test_window("two")
    assert_dimensions(manager, 200, 0, 600, 600)


@slice_config
def test_slice_last(manager):
    manager.test_window("one")
    assert_dimensions(manager, 200, 0, 600, 600)
    manager.test_window("slice")
    assert_dimensions(manager, 0, 0, 200, 600)


@slice_config
def test_slice_focus(manager):
    manager.test_window("one")
    assert_focused(manager, "one")
    two = manager.test_window("two")
    assert_focused(manager, "two")
    slice = manager.test_window("slice")
    assert_focused(manager, "slice")
    assert_focus_path(manager, "slice")
    manager.test_window("three")
    assert_focus_path(manager, "two", "one", "slice", "three")
    manager.kill_window(two)
    assert_focus_path(manager, "one", "slice", "three")
    manager.kill_window(slice)
    assert_focus_path(manager, "one", "three")
    slice = manager.test_window("slice")
    assert_focus_path(manager, "three", "one", "slice")


@slice_config
def test_all_slices(manager):
    manager.test_window("slice")  # left
    assert_dimensions(manager, 0, 0, 200, 600)
    manager.c.next_layout()  # right
    assert_dimensions(manager, 600, 0, 200, 600)
    manager.c.next_layout()  # top
    assert_dimensions(manager, 0, 0, 800, 200)
    manager.c.next_layout()  # bottom
    assert_dimensions(manager, 0, 400, 800, 200)
    manager.c.next_layout()  # left again
    manager.test_window("one")
    assert_dimensions(manager, 200, 0, 600, 600)
    manager.c.next_layout()  # right
    assert_dimensions(manager, 0, 0, 600, 600)
    manager.c.next_layout()  # top
    assert_dimensions(manager, 0, 200, 800, 400)
    manager.c.next_layout()  # bottom
    assert_dimensions(manager, 0, 0, 800, 400)


@slice_config
def test_command_propagation(manager):
    manager.test_window("slice")
    manager.test_window("one")
    manager.test_window("two")
    info = manager.c.layout.info()
    assert info["name"] == "slice"
    org_height = manager.c.window.info()["height"]
    manager.c.layout.toggle_split()
    assert manager.c.window.info()["height"] != org_height



# Copyright (c) 2011 Florian Mounier
# Copyright (c) 2012, 2014 Tycho Andersen
# Copyright (c) 2013 Craig Barnes
# Copyright (c) 2014 Sean Vig
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging

import pytest

import libqtile.bar
import libqtile.config
import libqtile.confreader
import libqtile.layout
import libqtile.log_utils
import libqtile.widget
from libqtile.command.base import CommandObject
from libqtile.command.interface import CommandError
from libqtile.confreader import Config
from libqtile.lazy import lazy
from test.conftest import dualmonitor


class CallConfig(Config):
    keys = [
        libqtile.config.Key(
            ["control"],
            "j",
            lazy.layout.down(),
        ),
        libqtile.config.Key(
            ["control"],
            "k",
            lazy.layout.up(),
        ),
    ]
    mouse = []
    groups = [
        libqtile.config.Group("a"),
        libqtile.config.Group("b"),
    ]
    layouts = [
        libqtile.layout.Stack(num_stacks=1),
        libqtile.layout.Max(),
    ]
    floating_layout = libqtile.resources.default_config.floating_layout
    screens = [
        libqtile.config.Screen(
            bottom=libqtile.bar.Bar(
                [
                    libqtile.widget.GroupBox(),
                ],
                20,
            ),
        )
    ]
    auto_fullscreen = True


call_config = pytest.mark.parametrize("manager", [CallConfig], indirect=True)


@call_config
def test_layout_filter(manager):
    manager.test_window("one")
    manager.test_window("two")
    assert manager.c.groups()["a"]["focus"] == "two"
    manager.c.simulate_keypress(["control"], "j")
    assert manager.c.groups()["a"]["focus"] == "one"
    manager.c.simulate_keypress(["control"], "k")
    assert manager.c.groups()["a"]["focus"] == "two"


class TestCommands(CommandObject):
    @staticmethod
    def cmd_one():
        pass

    def cmd_one_self(self):
        pass

    def cmd_two(self, a):
        pass

    def cmd_three(self, a, b=99):
        pass

    def _items(self, name):
        return None

    def _select(self, name, sel):
        return None


def test_doc():
    c = TestCommands()
    assert "one()" in c.cmd_doc("one")
    assert "one_self()" in c.cmd_doc("one_self")
    assert "two(a)" in c.cmd_doc("two")
    assert "three(a, b=99)" in c.cmd_doc("three")


def test_commands():
    c = TestCommands()
    assert len(c.cmd_commands()) == 9


def test_command():
    c = TestCommands()
    assert c.command("one")
    assert not c.command("nonexistent")


class ServerConfig(Config):
    auto_fullscreen = True
    keys = []
    mouse = []
    groups = [
        libqtile.config.Group("a"),
        libqtile.config.Group("b"),
        libqtile.config.Group("c"),
    ]
    layouts = [
        libqtile.layout.Stack(num_stacks=1),
        libqtile.layout.Stack(num_stacks=2),
        libqtile.layout.Stack(num_stacks=3),
    ]
    floating_layout = libqtile.resources.default_config.floating_layout
    screens = [
        libqtile.config.Screen(
            bottom=libqtile.bar.Bar(
                [
                    libqtile.widget.TextBox(name="one"),
                ],
                20,
            ),
        ),
        libqtile.config.Screen(
            bottom=libqtile.bar.Bar(
                [
                    libqtile.widget.TextBox(name="two"),
                ],
                20,
            ),
        ),
    ]


server_config = pytest.mark.parametrize("manager", [ServerConfig], indirect=True)


@server_config
def test_cmd_commands(manager):
    assert manager.c.commands()
    assert manager.c.layout.commands()
    assert manager.c.screen.bar["bottom"].commands()


@server_config
def test_call_unknown(manager):
    with pytest.raises(libqtile.command.client.SelectError, match="Not valid child or command"):
        manager.c.nonexistent

    manager.c.layout
    with pytest.raises(libqtile.command.client.SelectError, match="Not valid child or command"):
        manager.c.layout.nonexistent


@dualmonitor
@server_config
def test_items_qtile(manager):
    v = manager.c.items("group")
    assert v[0]
    assert sorted(v[1]) == ["a", "b", "c"]

    assert manager.c.items("layout") == (True, [0, 1, 2])

    v = manager.c.items("widget")
    assert not v[0]
    assert sorted(v[1]) == ["one", "two"]

    assert manager.c.items("bar") == (False, ["bottom"])
    t, lst = manager.c.items("window")
    assert t
    assert len(lst) == 2
    assert manager.c.window[lst[0]]
    assert manager.c.items("screen") == (True, [0, 1])


@dualmonitor
@server_config
def test_select_qtile(manager):
    assert manager.c.layout.info()["group"] == "a"
    assert len(manager.c.layout.info()["stacks"]) == 1
    assert len(manager.c.layout[2].info()["stacks"]) == 3
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        manager.c.layout[99]

    assert manager.c.group.info()["name"] == "a"
    assert manager.c.group["c"].info()["name"] == "c"
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        manager.c.group["nonexistent"]

    assert manager.c.widget["one"].info()["name"] == "one"
    with pytest.raises(CommandError, match="No object widget"):
        manager.c.widget.info()

    assert manager.c.bar["bottom"].info()["position"] == "bottom"

    manager.test_window("one")
    wid = manager.c.window.info()["id"]
    assert manager.c.window[wid].info()["id"] == wid

    assert manager.c.screen.info()["index"] == 0
    assert manager.c.screen[1].info()["index"] == 1
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        manager.c.screen[22]


@server_config
def test_items_group(manager):
    group = manager.c.group

    manager.test_window("test")
    wid = manager.c.window.info()["id"]

    assert group.items("window") == (True, [wid])
    assert group.items("layout") == (True, [0, 1, 2])
    assert group.items("screen") == (True, [])


@dualmonitor
@server_config
def test_select_group(manager):
    group = manager.c.group

    assert group.layout.info()["group"] == "a"
    assert len(group.layout.info()["stacks"]) == 1
    assert len(group.layout[2].info()["stacks"]) == 3

    with pytest.raises(CommandError):
        manager.c.group.window.info()
    manager.test_window("test")
    wid = manager.c.window.info()["id"]

    assert group.window.info()["id"] == wid
    assert group.window[wid].info()["id"] == wid
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        group.window["foo"]

    assert group.screen.info()["index"] == 0
    assert group["b"].screen.info()["index"] == 1
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        group.screen[0]


@server_config
def test_items_screen(manager):
    s = manager.c.screen
    assert s.items("layout") == (True, [0, 1, 2])

    manager.test_window("test")
    wid = manager.c.window.info()["id"]
    assert s.items("window") == (True, [wid])

    assert s.items("bar") == (False, ["bottom"])


@server_config
def test_select_screen(manager):
    screen = manager.c.screen
    assert screen.layout.info()["group"] == "a"
    assert len(screen.layout.info()["stacks"]) == 1
    assert len(screen.layout[2].info()["stacks"]) == 3

    with pytest.raises(CommandError):
        manager.c.window.info()

    manager.test_window("test")
    wid = manager.c.window.info()["id"]
    assert screen.window.info()["id"] == wid
    assert screen.window[wid].info()["id"] == wid

    with pytest.raises(CommandError, match="No object"):
        screen.bar.info()
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        screen.bar["top"]

    assert screen.bar["bottom"].info()["position"] == "bottom"


@server_config
def test_items_bar(manager):
    assert manager.c.bar["bottom"].items("screen") == (True, [])


@dualmonitor
@server_config
def test_select_bar(manager):
    assert manager.c.screen[1].bar["bottom"].screen.info()["index"] == 1
    b = manager.c.bar
    assert b["bottom"].screen.info()["index"] == 0
    with pytest.raises(CommandError):
        b.screen.info()


@server_config
def test_items_layout(manager):
    assert manager.c.layout.items("screen") == (True, [])
    assert manager.c.layout.items("group") == (True, [])


@server_config
def test_select_layout(manager):
    layout = manager.c.layout

    assert layout.screen.info()["index"] == 0
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        layout.screen[0]

    assert layout.group.info()["name"] == "a"
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        layout.group["a"]


@dualmonitor
@server_config
def test_items_window(manager):
    manager.test_window("test")
    window = manager.c.window
    window.info()["id"]

    assert window.items("group") == (True, [])
    assert window.items("layout") == (True, [0, 1, 2])
    assert window.items("screen") == (True, [])


@dualmonitor
@server_config
def test_select_window(manager):
    manager.test_window("test")
    window = manager.c.window
    window.info()["id"]

    assert window.group.info()["name"] == "a"
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        window.group["a"]

    assert len(window.layout.info()["stacks"]) == 1
    assert len(window.layout[1].info()["stacks"]) == 2

    assert window.screen.info()["index"] == 0
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        window.screen[0]


@server_config
def test_items_widget(manager):
    assert manager.c.widget["one"].items("bar") == (True, [])


@server_config
def test_select_widget(manager):
    widget = manager.c.widget["one"]
    assert widget.bar.info()["position"] == "bottom"
    with pytest.raises(libqtile.command.client.SelectError, match="Item not available in object"):
        widget.bar["bottom"]


def test_core_node(manager, backend_name):
    assert manager.c.core.info()["backend"] == backend_name


def test_lazy_arguments(manager_nospawn):

    # Decorated function to be bound to key presses
    @lazy.function
    def test_func(qtile, value, multiplier=1):
        qtile.test_func_output = value * multiplier

    config = ServerConfig
    config.keys = [
        libqtile.config.Key(
            ["control"],
            "j",
            test_func(10),
        ),
        libqtile.config.Key(["control"], "k", test_func(5, multiplier=100)),
    ]

    manager_nospawn.start(config)

    manager_nospawn.c.simulate_keypress(["control"], "j")
    _, val = manager_nospawn.c.eval("self.test_func_output")
    assert val == "10"

    manager_nospawn.c.simulate_keypress(["control"], "k")
    _, val = manager_nospawn.c.eval("self.test_func_output")
    assert val == "500"


def test_deprecated_modules(caplog):
    libqtile.log_utils.init_log(logging.WARNING, log_path=None, log_color=False)

    from libqtile.command_client import InteractiveCommandClient  # noqa: F401

    assert caplog.record_tuples == [
        (
            "libqtile",
            logging.WARNING,
            "libqtile.command_client is deprecated. It has been moved to libqtile.command.client",
        )
    ]

    caplog.clear()

    from libqtile.command_graph import CommandGraphNode  # noqa: F401

    assert caplog.record_tuples == [
        (
            "libqtile",
            logging.WARNING,
            "libqtile.command_graph is deprecated. It has been moved to libqtile.command.graph",
        )
    ]

    caplog.clear()

    from libqtile.command_interface import CommandInterface  # noqa: F401

    assert caplog.record_tuples == [
        (
            "libqtile",
            logging.WARNING,
            "libqtile.command_interface is deprecated. It has been moved to libqtile.command.interface",
        )
    ]

    caplog.clear()

    from libqtile.command_object import CommandObject  # noqa: F401

    assert caplog.record_tuples == [
        (
            "libqtile",
            logging.WARNING,
            "libqtile.command_object is deprecated. It has been moved to libqtile.command.base.",
        )
    ]



# Copyright (c) 2021 Matt Colligan
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import os
from typing import TYPE_CHECKING

from pywayland.protocol.wayland.wl_output import WlOutput
from wlroots.util.box import Box
from wlroots.util.clock import Timespec
from wlroots.util.region import PixmanRegion32
from wlroots.wlr_types import Matrix
from wlroots.wlr_types import Output as wlrOutput
from wlroots.wlr_types import OutputDamage
from wlroots.wlr_types.layer_shell_v1 import LayerShellV1Layer, LayerSurfaceV1Anchor

from libqtile.backend.wayland.window import Internal, LayerStatic
from libqtile.backend.wayland.wlrq import HasListeners
from libqtile.log_utils import logger

if TYPE_CHECKING:
    from typing import Any, Sequence

    from pywayland.server import Listener
    from wlroots.wlr_types import Surface, Texture

    from libqtile.backend.wayland.core import Core
    from libqtile.backend.wayland.window import WindowType
    from libqtile.backend.wayland.wlrq import Dnd
    from libqtile.config import Screen

no_transform = WlOutput.transform.normal


class Output(HasListeners):
    def __init__(self, core: Core, wlr_output: wlrOutput):
        self.core = core
        self.renderer = core.renderer
        self.wlr_output = wlr_output
        wlr_output.data = self
        self.output_layout = self.core.output_layout
        self._damage: OutputDamage = OutputDamage(wlr_output)
        self.wallpaper: Texture | None = None
        self.x, self.y = self.output_layout.output_coords(wlr_output)

        self.add_listener(wlr_output.destroy_event, self._on_destroy)
        self.add_listener(self._damage.frame_event, self._on_frame)

        # The layers enum indexes into this list to get a list of surfaces
        self.layers: list[list[LayerStatic]] = [[] for _ in range(len(LayerShellV1Layer))]

        # This is run during tests, when we want to fix the output's geometry
        if wlr_output.is_headless and "PYTEST_CURRENT_TEST" in os.environ:
            assert len(core.outputs) < 2, "This should not be reached"
            if not core.outputs:
                # First test output
                wlr_output.set_custom_mode(800, 600, 0)
            else:
                # Second test output
                wlr_output.set_custom_mode(640, 480, 0)
            wlr_output.commit()

    def finalize(self) -> None:
        self.core.remove_output(self)
        self.finalize_listeners()

    @property
    def screen(self) -> Screen:
        assert self.core.qtile is not None

        if len(self.core.qtile.screens) > 1:
            x, y, w, h = self.get_geometry()
            for screen in self.core.qtile.screens:
                if screen.x == x and screen.y == y:
                    if screen.width == w and screen.height == h:
                        return screen
        return self.core.qtile.current_screen

    def _on_destroy(self, _listener: Listener, _data: Any) -> None:
        logger.debug("Signal: output destroy")
        self.finalize()

    def _on_frame(self, _listener: Listener, _data: Any) -> None:
        with PixmanRegion32() as damage:
            try:
                if not self._damage.attach_render(damage):
                    # No new frame needed.
                    self.wlr_output.rollback()
                    return
            except RuntimeError:
                # Failed to attach render; skip.
                return

            with self.wlr_output as wlr_output:
                if not damage.not_empty():
                    # No damage, only buffer swap needed.
                    return

                now = Timespec.get_monotonic_time()
                scale = wlr_output.scale
                transform_matrix = wlr_output.transform_matrix

                with self.renderer.render(
                    wlr_output._ptr.width, wlr_output._ptr.height
                ) as renderer:

                    if self.wallpaper:
                        width, height = wlr_output.effective_resolution()
                        box = Box(0, 0, int(width * scale), int(height * scale))
                        matrix = Matrix.project_box(box, no_transform, 0, transform_matrix)
                        renderer.render_texture_with_matrix(self.wallpaper, matrix, 1)
                    else:
                        renderer.clear([0, 0, 0, 1])

                    mapped: Sequence[WindowType] = (
                        self.layers[LayerShellV1Layer.BACKGROUND]
                        + self.layers[LayerShellV1Layer.BOTTOM]
                        + self.core.mapped_windows  # type: ignore
                        + self.layers[LayerShellV1Layer.TOP]
                        + self.layers[LayerShellV1Layer.OVERLAY]
                    )

                    for window in mapped:
                        if isinstance(window, Internal):
                            box = Box(
                                int((window.x - self.x) * scale),
                                int((window.y - self.y) * scale),
                                int(window.width * scale),
                                int(window.height * scale),
                            )
                            matrix = Matrix.project_box(box, no_transform, 0, transform_matrix)
                            renderer.render_texture_with_matrix(
                                window.texture, matrix, window.opacity
                            )
                        else:
                            rdata = (
                                now,
                                window,
                                window.x - self.x,  # layout coordinates -> output coordinates
                                window.y - self.y,
                                window.opacity,
                                scale,
                                transform_matrix,
                            )
                            window.surface.for_each_surface(self._render_surface, rdata)

                    if self.core.live_dnd:
                        self._render_dnd_icon(self.core.live_dnd, now, scale, transform_matrix)

                    wlr_output.render_software_cursors(damage=damage)

    def _render_surface(self, surface: Surface, sx: int, sy: int, rdata: tuple) -> None:
        texture = surface.get_texture()
        if texture is None:
            return

        now, window, wx, wy, opacity, scale, transform_matrix = rdata
        x = (wx + sx) * scale
        y = (wy + sy) * scale
        width = surface.current.width * scale
        height = surface.current.height * scale

        if window.borderwidth:
            bw = int(window.borderwidth * scale)

            if surface == window.surface.surface:
                outer_w = width + bw * 2
                outer_h = height + bw * 2
                num = len(window.bordercolor)
                bws = [bw // num] * num
                for i in range(bw % num):
                    bws[i] += 1
                coord = 0
                for i, bc in enumerate(window.bordercolor):
                    border = Box(
                        int(x + coord),
                        int(y + coord),
                        int(outer_w - coord * 2),
                        int(bws[i]),
                    )
                    self.renderer.render_rect(border, bc, transform_matrix)  # Top border
                    border.y = int(y + outer_h - bws[i] - coord)
                    self.renderer.render_rect(border, bc, transform_matrix)  # Bottom border
                    border.y = int(y + coord)
                    border.width = int(bws[i])
                    border.height = int(outer_h - coord * 2)
                    self.renderer.render_rect(border, bc, transform_matrix)  # Left border
                    border.x = int(x + outer_w - bws[i] - coord)
                    self.renderer.render_rect(border, bc, transform_matrix)  # Right border
                    coord += bws[i]

            x += bw
            y += bw

        box = Box(
            int(x),
            int(y),
            int(width),
            int(height),
        )

        inverse = wlrOutput.transform_invert(surface.current.transform)
        matrix = Matrix.project_box(box, inverse, 0, transform_matrix)
        self.renderer.render_texture_with_matrix(texture, matrix, opacity)
        surface.send_frame_done(now)

    def _render_dnd_icon(
        self, dnd: Dnd, now: Timespec, scale: float, transform_matrix: Matrix
    ) -> None:
        """Render the drag-n-drop icon if there is one."""
        icon = dnd.wlr_drag.icon
        if icon.mapped:
            texture = icon.surface.get_texture()
            if texture:
                box = Box(
                    int((dnd.x - self.x) * scale),
                    int((dnd.y - self.y) * scale),
                    int(icon.surface.current.width * scale),
                    int(icon.surface.current.height * scale),
                )
                inverse = wlrOutput.transform_invert(icon.surface.current.transform)
                matrix = Matrix.project_box(box, inverse, 0, transform_matrix)
                self.renderer.render_texture_with_matrix(texture, matrix, 1)
                icon.surface.send_frame_done(now)

    def get_geometry(self) -> tuple[int, int, int, int]:
        width, height = self.wlr_output.effective_resolution()
        return int(self.x), int(self.y), width, height

    def organise_layers(self) -> None:
        """Organise the positioning of layer shell surfaces."""
        logger.debug("Output: organising layers")
        ow, oh = self.wlr_output.effective_resolution()

        for layer in self.layers:
            for win in layer:
                state = win.surface.current
                margin = state.margin
                ww = win.desired_width = state.desired_width
                wh = win.desired_height = state.desired_height

                # Horizontal axis
                if (state.anchor & LayerSurfaceV1Anchor.HORIZONTAL) and ww == 0:
                    x = margin.left
                    ww = ow - margin.left - margin.right
                elif state.anchor & LayerSurfaceV1Anchor.LEFT:
                    x = margin.left
                elif state.anchor & LayerSurfaceV1Anchor.RIGHT:
                    x = ow - ww - margin.right
                else:
                    x = int(ow / 2 - ww / 2)

                # Vertical axis
                if (state.anchor & LayerSurfaceV1Anchor.VERTICAL) and wh == 0:
                    y = margin.top
                    wh = oh - margin.top - margin.bottom
                elif state.anchor & LayerSurfaceV1Anchor.TOP:
                    y = margin.top
                elif state.anchor & LayerSurfaceV1Anchor.BOTTOM:
                    y = oh - wh - margin.bottom
                else:
                    y = int(oh / 2 - wh / 2)

                if ww <= 0 or wh <= 0:
                    win.kill()
                    continue

                if 0 < state.exclusive_zone:
                    # Reserve space if:
                    #    - layer is anchored to an edge and both perpendicular edges, or
                    #    - layer is anchored to a single edge only.
                    space = [0, 0, 0, 0]

                    if state.anchor & LayerSurfaceV1Anchor.HORIZONTAL:
                        if state.anchor & LayerSurfaceV1Anchor.TOP:
                            space[2] = state.exclusive_zone
                        elif state.anchor & LayerSurfaceV1Anchor.BOTTOM:
                            space[3] = state.exclusive_zone
                    elif state.anchor & LayerSurfaceV1Anchor.VERTICAL:
                        if state.anchor & LayerSurfaceV1Anchor.LEFT:
                            space[0] = state.exclusive_zone
                        elif state.anchor & LayerSurfaceV1Anchor.RIGHT:
                            space[1] = state.exclusive_zone
                    else:
                        # Single edge only
                        if state.anchor == LayerSurfaceV1Anchor.TOP:
                            space[2] = state.exclusive_zone
                        elif state.anchor == LayerSurfaceV1Anchor.BOTTOM:
                            space[3] = state.exclusive_zone
                        if state.anchor == LayerSurfaceV1Anchor.LEFT:
                            space[0] = state.exclusive_zone
                        elif state.anchor == LayerSurfaceV1Anchor.RIGHT:
                            space[1] = state.exclusive_zone

                    to_reserve: tuple[int, int, int, int] = tuple(space)  # type: ignore
                    if win.reserved_space != to_reserve:
                        # Don't reserve more space if it's already been reserved
                        assert self.core.qtile is not None
                        self.core.qtile.reserve_space(to_reserve, self.screen)
                        win.reserved_space = to_reserve

                win.place(int(x + self.x), int(y + self.y), int(ww), int(wh), 0, None)

        self.core.stack_windows()

    def contains(self, rect: WindowType | Dnd) -> bool:
        """Returns whether the given window is visible on this output."""
        if rect.x + rect.width < self.x:
            return False
        if rect.y + rect.height < self.y:
            return False

        ow, oh = self.wlr_output.effective_resolution()
        if self.x + ow < rect.x:
            return False
        if self.y + oh < rect.y:
            return False

        return True

    def damage(self) -> None:
        """Damage this output so it gets re-rendered."""
        if self.wlr_output.enabled:
            self._damage.add_whole()



# -*- coding: utf-8 -*-
import logging

from libqtile.widget._pulse_audio import ffi, lib
from libqtile.widget.volume import Volume

log = logging.getLogger(__name__)


@ffi.def_extern()
def qtile_pa_context_changed(context, userdata):
    """callback for connecting status update"""
    widget = ffi.from_handle(userdata)
    widget.on_connection_change(context)


@ffi.def_extern()
def qtile_on_sink_info(context, info, eol, userdata):
    """callback for each output sink that pulseaudio server has"""
    widget = ffi.from_handle(userdata)
    widget.on_sink_info(info, eol)


@ffi.def_extern()
def qtile_on_server_info(context, info, userdata):
    """callback with a pulseaudio server info"""
    widget = ffi.from_handle(userdata)
    widget.on_server_info(info)


@ffi.def_extern()
def qtile_on_sink_update(context, event_type, sink_index, userdata):
    """callback for updates made to sinks"""
    widget = ffi.from_handle(userdata)
    widget.on_sink_update(event_type, sink_index)


class PulseVolume(Volume):
    defaults = [
        ("limit_max_volume", False, "Limit maximum volume to 100%"),
    ]

    def __init__(self, **config):
        Volume.__init__(self, **config)
        self.add_defaults(PulseVolume.defaults)

        self.connected = None
        self._subscribed = False
        self.default_sink_name = None
        self.default_sink = None
        self.handle = ffi.new_handle(self)
        self.client_name = ffi.new("char[]", b"Qtile-pulse")

        self.connect()

    def finalize(self):
        lib.pa_context_disconnect(self.context)
        lib.pa_mainloop_quit(self.loop, 1)
        lib.pa_context_unref(self.context)
        lib.pa_mainloop_free(self.loop)
        Volume.finalize(self)

    def connect(self):
        """
        issue a connection to pulse audio server. result of a connection
        would be passed to `on_connection_change` method
        """
        # create a loop and api entry point
        self.loop = lib.pa_mainloop_new()
        self.api = lib.pa_mainloop_get_api(self.loop)
        # create context (e.g. connection)
        self.context = lib.pa_context_new(self.api, self.client_name)
        lib.pa_context_connect(self.context, ffi.NULL, 0, ffi.NULL)
        lib.pa_context_set_state_callback(self.context, lib.qtile_pa_context_changed, self.handle)

    def on_connection_change(self, context):
        """a callback from pulse lib indicating connection status"""
        state = lib.pa_context_get_state(context)
        if state == lib.PA_CONTEXT_READY:
            # ready
            self.connected = True
            # once onnection is established we need to get server information
            self.timeout_add(0.1, self.get_server_info)
            log.debug("Connection to pulseaudio ready")
        elif state == lib.PA_CONTEXT_FAILED:
            # failed to connect
            self.connected = False
            self._subscribed = False
            log.warning("Failed to connect to pulseaudio, retrying in 10s")
            self.timeout_add(10, self.connect)
        elif state == lib.PA_CONTEXT_TERMINATED:
            # done
            self.connected = False
            self._subscribed = False
            log.debug("Connection to pulseaudio terminated cleanly")
        elif state == lib.PA_CONTEXT_UNCONNECTED:
            self.connected = False
            self._subscribed = False
            log.warning("Disconnected from pulseaudio")

    def get_server_info(self):
        lib.pa_context_get_server_info(self.context, lib.qtile_on_server_info, self.handle)

    def on_server_info(self, info):
        self.default_sink_name = ffi.string(info.default_sink_name).decode("utf-8")
        self.timeout_add(0.1, self.get_sinks)

    def get_sinks(self):
        lib.pa_context_get_sink_info_list(self.context, lib.qtile_on_sink_info, self.handle)

    def on_sink_info(self, sink, eol):
        if eol:  # dont operate on sink in case its an eol callback
            if not self._subscribed:
                self.timeout_add(0.1, self.subscribe_to_sink_events)
            return
        name = ffi.string(sink.name).decode("utf-8")
        if name == self.default_sink_name:
            self.default_sink = {
                "name": name,
                "description": ffi.string(sink.description).decode("utf-8"),
                "index": int(sink.index),
                "base_volume": sink.base_volume,
                "muted": bool(sink.mute),
                "channels": sink.volume.channels,
                "values": list(sink.volume.values),
            }
            self.update()

    def subscribe_to_sink_events(self):
        op = lib.pa_context_subscribe(
            self.context,
            lib.PA_SUBSCRIPTION_MASK_SINK,
            ffi.NULL,
            ffi.NULL,
        )
        self.wait_for_operation(op)
        lib.pa_context_set_subscribe_callback(self.context, lib.qtile_on_sink_update, self.handle)
        self._subscribed = True

    def on_sink_update(self, event_type, sink_index):
        self.timeout_add(0.1, self.get_sinks)

    def wait_for_operation(self, op):
        state = lib.pa_operation_get_state(op)
        while state == lib.PA_OPERATION_RUNNING:
            lib.pa_mainloop_iterate(self.loop, 0, ffi.NULL)
            state = lib.pa_operation_get_state(op)

    def change_volume(self, volume):
        """
        order pulseaudio to apply new volume
        """
        # store new volume to "speed up" widget update so that we don't have
        # to wait a callback from pulseaudio
        self.default_sink["values"] = list(volume.values)
        op = lib.pa_context_set_sink_volume_by_index(
            self.context, self.default_sink["index"], volume, ffi.NULL, ffi.NULL
        )
        if op:
            self.wait_for_operation(op)

    def cmd_mute(self):
        op = lib.pa_context_set_sink_mute_by_index(
            self.context,
            self.default_sink["index"],
            not self.default_sink["muted"],
            ffi.NULL,
            ffi.NULL,
        )
        if op:
            self.wait_for_operation(op)

    def cmd_increase_vol(self, value=None):
        if value is None:
            value = self.step
        base = self.default_sink["base_volume"]
        volume = ffi.new(
            "pa_cvolume *",
            {
                "channels": self.default_sink["channels"],
                "values": self.default_sink["values"],
            },
        )
        lib.pa_cvolume_inc(
            volume,
            int(value * base / 100),
        )
        # check that we dont go over 100% in case its set in config
        if self.limit_max_volume:
            volume.values = [(i if i <= base else base) for i in volume.values]
        self.change_volume(volume)

    def cmd_decrease_vol(self, value=None):
        if value is None:
            value = self.step
        volume_level = int(value * self.default_sink["base_volume"] / 100)
        if not volume_level and max(self.default_sink["values"]) == 0:
            # can't be lower than zero
            return
        volume = ffi.new(
            "pa_cvolume *",
            {
                "channels": self.default_sink["channels"],
                "values": self.default_sink["values"],
            },
        )
        lib.pa_cvolume_dec(volume, volume_level)
        self.change_volume(volume)

    def button_press(self, x, y, button):
        Volume.button_press(self, x, y, button)
        self.poll()

    def poll(self):
        lib.pa_mainloop_iterate(self.loop, 0, ffi.NULL)
        self.update()

    def update(self):
        """
        same method as in Volume widgets except that here we don't need to
        manually re-schedule update
        """
        vol = self.get_volume()
        if vol != self.volume:
            self.volume = vol
            # Update the underlying canvas size before actually attempting
            # to figure out how big it is and draw it.
            self._update_drawer()
            self.bar.draw()

    def get_volume(self):
        if self.default_sink:
            if self.default_sink["muted"]:
                return -1
            base = self.default_sink["base_volume"]
            if not base:
                return -1
            current = max(self.default_sink["values"])
            return round(current * 100 / base)
        return -1

    def timer_setup(self):
        self.poll()
        if self.update_interval is not None:
            self.timeout_add(self.update_interval, self.timer_setup)
        if self.theme_path:
            self.setup_images()



import pytest

from libqtile.backend import get_core
from libqtile.backend.x11 import core
from test.test_manager import ManagerConfig


def test_get_core_x11(display):
    get_core("x11", display).finalize()


def test_keys(display):
    assert "a" in core.get_keys()
    assert "shift" in core.get_modifiers()


def test_no_two_qtiles(xmanager):
    with pytest.raises(core.ExistingWMException):
        core.Core(xmanager.display).finalize()


def test_color_pixel(xmanager):
    (success, e) = xmanager.c.eval('self.core.conn.color_pixel("ffffff")')
    assert success, e


@pytest.mark.parametrize("xmanager", [ManagerConfig], indirect=True)
def test_net_client_list(xmanager, conn):
    def assert_clients(number):
        clients = conn.default_screen.root.get_property("_NET_CLIENT_LIST", unpack=int)
        assert len(clients) == number

    # ManagerConfig has a Bar, which should not appear in _NET_CLIENT_LIST
    assert_clients(0)
    one = xmanager.test_window("one")
    assert_clients(1)
    two = xmanager.test_window("two")
    xmanager.c.window.toggle_minimize()
    three = xmanager.test_window("three")
    xmanager.c.screen.next_group()
    assert_clients(3)  # Minimized windows and windows on other groups are included
    xmanager.kill_window(one)
    xmanager.c.screen.next_group()
    assert_clients(2)
    xmanager.kill_window(three)
    assert_clients(1)
    xmanager.c.screen.next_group()
    one = xmanager.test_window("one")
    assert_clients(2)
    xmanager.c.window.static()  # Static windows are not included
    assert_clients(1)
    xmanager.kill_window(two)
    assert_clients(0)



# -*- coding: utf-8 -*-
# Copyright (c) 2015 Jörg Thalheim (Mic92)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import psutil

from libqtile.widget import base

__all__ = ["Memory"]


class Memory(base.ThreadPoolText):
    """Displays memory/swap usage

    MemUsed: Returns memory in use
    MemTotal: Returns total amount of memory
    MemFree: Returns amount of memory free
    MemPercent: Returns memory in use as a percentage
    Buffers: Returns buffer amount
    Active: Returns active memory
    Inactive: Returns inactive memory
    Shmem: Returns shared memory
    SwapTotal: Returns total amount of swap
    SwapFree: Returns amount of swap free
    SwapUsed: Returns amount of swap in use
    SwapPercent: Returns swap in use as a percentage


    Widget requirements: psutil_.

    .. _psutil: https://pypi.org/project/psutil/
    """

    defaults = [
        ("format", "{MemUsed: .0f}{mm}/{MemTotal: .0f}{mm}", "Formatting for field names."),
        ("update_interval", 1.0, "Update interval for the Memory"),
        ("measure_mem", "M", "Measurement for Memory (G, M, K, B)"),
        ("measure_swap", "M", "Measurement for Swap (G, M, K, B)"),
    ]

    measures = {"G": 1024 * 1024 * 1024, "M": 1024 * 1024, "K": 1024, "B": 1}

    def __init__(self, **config):
        super().__init__("", **config)
        self.add_defaults(Memory.defaults)
        self.calc_mem = self.measures[self.measure_mem]
        self.calc_swap = self.measures[self.measure_swap]

    def poll(self):
        mem = psutil.virtual_memory()
        swap = psutil.swap_memory()
        val = {}
        val["MemUsed"] = mem.used / self.calc_mem
        val["MemTotal"] = mem.total / self.calc_mem
        val["MemFree"] = mem.free / self.calc_mem
        val["MemPercent"] = mem.percent
        val["Buffers"] = mem.buffers / self.calc_mem
        val["Active"] = mem.active / self.calc_mem
        val["Inactive"] = mem.inactive / self.calc_mem
        val["Shmem"] = mem.shared / self.calc_mem
        val["SwapTotal"] = swap.total / self.calc_swap
        val["SwapFree"] = swap.free / self.calc_swap
        val["SwapUsed"] = swap.used / self.calc_swap
        val["SwapPercent"] = swap.percent
        val["mm"] = self.measure_mem
        val["ms"] = self.measure_swap
        return self.format.format(**val)



import importlib

from libqtile.utils import QtileError

CORES = [
    "wayland",
    "x11",
]


def get_core(backend, *args):
    if backend not in CORES:
        raise QtileError(f"Backend {backend} does not exist")

    return importlib.import_module(f"libqtile.backend.{backend}.core").Core(*args)



#!/usr/bin/env python3

# Copyright (c) 2008 Aldo Cortesi
# Copyright (c) 2011 Mounier Florian
# Copyright (c) 2012 dmpayton
# Copyright (c) 2014 Sean Vig
# Copyright (c) 2014 roger
# Copyright (c) 2014 Pedro Algarvio
# Copyright (c) 2014-2015 Tycho Andersen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import textwrap

from setuptools import setup
from setuptools.command.install import install


class CheckCairoXcb(install):
    def cairo_xcb_check(self):
        try:
            from cairocffi import cairo
            cairo.cairo_xcb_surface_create
            return True
        except AttributeError:
            return False

    def finalize_options(self):
        if not self.cairo_xcb_check():

            print(textwrap.dedent("""

            It looks like your cairocffi was not built with xcffib support.  To fix this:

              - Ensure a recent xcffib is installed (pip install 'xcffib>=0.5.0')
              - The pip cache is cleared (remove ~/.cache/pip, if it exists)
              - Reinstall cairocffi, either:

                  pip install --no-deps --ignore-installed cairocffi

                or

                  pip uninstall cairocffi && pip install cairocffi
            """))

            sys.exit(1)
        install.finalize_options(self)


def get_cffi_modules():
    cffi_modules = [
        'libqtile/pango_ffi_build.py:pango_ffi',
        'libqtile/backend/x11/xcursors_ffi_build.py:xcursors_ffi',
    ]
    try:
        from cffi.error import PkgConfigError
        from cffi.pkgconfig import call
    except ImportError:
        # technically all ffi defined above wont be built
        print('CFFI package is missing')
    else:
        try:
            call('libpulse', '--libs')
        except PkgConfigError:
            print('Failed to find pulseaudio headers. '
                  'PulseVolume widget will be unavailable')
        else:
            cffi_modules.append(
                'libqtile/widget/pulseaudio_ffi.py:pulseaudio_ffi'
            )
    try:
        import wlroots.ffi_build
        cffi_modules.append(
            'libqtile/backend/wayland/libinput_ffi_build.py:libinput_ffi',
        )
    except ImportError:
        print(
            "Failed to find pywlroots. "
            "Wayland backend libinput configuration will be unavailable."
        )
        pass
    return cffi_modules


setup(
    cmdclass={'install': CheckCairoXcb},
    use_scm_version=True,
    cffi_modules=get_cffi_modules(),
    include_package_data=True,
)



"""
test_images.py contains unittests for libqtile.images.Img
and its supporting code.
"""
import os
from glob import glob
from os import path

import cairocffi
import cairocffi.pixbuf
import pytest

from libqtile import images

TEST_DIR = path.dirname(os.path.abspath(__file__))
DATA_DIR = path.join(TEST_DIR, "data")
PNGS = glob(path.join(DATA_DIR, "*", "*.png"))
SVGS = glob(path.join(DATA_DIR, "*", "*.svg"))
ALL_IMAGES = glob(path.join(DATA_DIR, "*", "*"))


@pytest.fixture(
    scope="function",
    params=ALL_IMAGES,
)
def path_n_bytes_image(request):
    fpath = request.param
    with open(fpath, "rb") as fobj:
        bobj = fobj.read()
    return fpath, bobj


@pytest.fixture(
    scope="function",
    params=PNGS,
)
def path_n_bytes_image_pngs(request):
    fpath = request.param
    with open(fpath, "rb") as fobj:
        bobj = fobj.read()
    return fpath, bobj


@pytest.fixture(scope="function")
def png_img():
    return images.Img.from_path(PNGS[0])


def test_get_cairo_surface(path_n_bytes_image):
    path, bytes_image = path_n_bytes_image
    surf_info = images.get_cairo_surface(bytes_image)
    assert isinstance(surf_info.surface, cairocffi.ImageSurface)
    assert path.split(".")[-1].lower() == surf_info.file_type


def test_get_cairo_surface_bad_input():
    with pytest.raises(cairocffi.pixbuf.ImageLoadingError):
        images.get_cairo_surface(b"asdfasfdi3")


def assert_approx_equal(vec0, vec1):
    approx = pytest.approx
    for val0, val1 in zip(vec0, vec1):
        assert val0 == approx(val1)


class TestImg:
    def test_init(self, path_n_bytes_image):
        path, bytes_image = path_n_bytes_image
        img = images.Img(bytes_image)
        assert isinstance(img.surface, cairocffi.ImageSurface)
        del img.surface
        assert isinstance(img.surface, cairocffi.ImageSurface)

    def test_from_path(self, path_n_bytes_image):
        path, bytes_image = path_n_bytes_image
        img = images.Img(bytes_image)
        assert isinstance(img.surface, cairocffi.ImageSurface)
        img2 = img.from_path(path)
        assert img == img2
        img2.theta = 90.0
        assert img != img2
        img2.theta = 0.0
        assert img == img2

    def test_setting(self, png_img):
        img = png_img
        width0, height0 = img.default_size
        pat0 = img.pattern
        img.width = width0 + 3
        assert pat0 != img.pattern
        assert img.width == (width0 + 3)
        pat1 = img.pattern
        img.height = height0 + 7
        assert img.height == (height0 + 7)
        assert img.pattern != pat0
        assert img.pattern != pat1
        pat2 = img.pattern
        img.theta = -35.0
        assert img.pattern != pat0
        assert img.pattern != pat1
        assert img.pattern != pat2
        assert img.theta == pytest.approx(-35.0)

    def test_equality(self, png_img):
        width0, height0 = png_img.default_size
        png_img2 = images.Img.from_path(png_img.path)
        assert png_img == png_img2
        png_img.width = width0 * 2
        png_img2.height = width0 * 2
        assert png_img != png_img2

    def test_setting_negative_size(self, png_img):
        png_img.width = -90
        assert png_img.width == 1
        png_img.height = 0
        assert png_img.height == 1

    def test_pattern(self, path_n_bytes_image):
        path, bytes_image = path_n_bytes_image
        img = images.Img(bytes_image)
        assert isinstance(img.pattern, cairocffi.SurfacePattern)

    def test_pattern_resize(self, path_n_bytes_image_pngs):
        path, bytes_image = path_n_bytes_image_pngs
        img = images.Img.from_path(path)
        assert isinstance(img.pattern, cairocffi.SurfacePattern)
        t_matrix = img.pattern.get_matrix().as_tuple()
        assert_approx_equal(t_matrix, (1.0, 0.0, 0.0, 1.0))
        img.width = 2.0 * img.default_size.width
        t_matrix = img.pattern.get_matrix().as_tuple()
        assert_approx_equal(t_matrix, (0.5, 0.0, 0.0, 1.0))
        img.height = 3.0 * img.default_size.height
        t_matrix = img.pattern.get_matrix().as_tuple()
        assert_approx_equal(t_matrix, (0.5, 0.0, 0.0, 1.0 / 3.0))

    def test_pattern_rotate(self, path_n_bytes_image):
        path, bytes_image = path_n_bytes_image
        img = images.Img(bytes_image)
        img.theta = 90.0
        assert img.theta == 90.0
        t_matrix = img.pattern.get_matrix().as_tuple()
        assert_approx_equal(t_matrix, (0.0, 1.0, -1.0, 0.0))
        img.theta = 45.0
        t_matrix = img.pattern.get_matrix().as_tuple()
        from math import sqrt

        s2o2 = sqrt(2) / 2.0
        assert_approx_equal(t_matrix, (s2o2, s2o2, -s2o2, s2o2))
        del img.theta
        assert img.theta == pytest.approx(0.0)


class TestImgScale:
    def test_scale(self, png_img):
        size = png_img.default_size
        png_img.scale(2, 3)
        assert png_img.width == 2 * size.width
        assert png_img.height == 3 * size.height

    def test_scale_rounding(self, png_img):
        size = png_img.default_size
        png_img.scale(1.99999, 2.99999)
        assert png_img.width == 2 * size.width
        assert png_img.height == 3 * size.height

    def test_scale_width_lock(self, png_img):
        size = png_img.default_size
        png_img.scale(width_factor=10, lock_aspect_ratio=True)
        assert png_img.width == 10 * size.width
        assert png_img.height == 10 * size.height

    def test_scale_height_lock(self, png_img):
        size = png_img.default_size
        png_img.scale(height_factor=11, lock_aspect_ratio=True)
        assert png_img.height == 11 * size.height
        assert png_img.width == 11 * size.width

    def test_scale_fail_lock(self, png_img):
        with pytest.raises(ValueError):
            png_img.scale(0.5, 4.0, lock_aspect_ratio=True)

    def test_scale_fail(self, png_img):
        with pytest.raises(ValueError):
            png_img.scale()


class TestImgResize:
    def test_resize(self, png_img):
        png_img.resize(100, 100)
        assert png_img.width == 100
        assert png_img.height == 100

    def test_resize_width(self, png_img):
        size = png_img.default_size
        ratio = size.width / size.height
        png_img.resize(width=40)
        assert png_img.width == 40
        assert (png_img.width / png_img.height) == pytest.approx(ratio)

    def test_resize_height(self, png_img):
        size = png_img.default_size
        ratio = size.width / size.height
        png_img.resize(height=10)
        assert png_img.height == 10
        assert (png_img.width / png_img.height) == pytest.approx(ratio)


class TestLoader:
    @pytest.fixture(scope="function")
    def loader(self):
        png_dir = path.join(DATA_DIR, "png")
        svg_dir = path.join(DATA_DIR, "svg")
        return images.Loader(svg_dir, png_dir)

    def test_audio_volume_muted(self, loader):
        name = "audio-volume-muted"
        result = loader(name)
        assert isinstance(result[name], images.Img)
        assert result[name].path.endswith(".svg")

    def test_audio_volume_muted_png(self, loader):
        name = "audio-volume-muted.png"
        result = loader(name)
        assert isinstance(result[name], images.Img)
        assert result[name].path.endswith(".png")

    def test_load_file_missing(self, loader):
        names = ("audio-asdlfjasdvolume-muted", "audio-volume-muted")
        with pytest.raises(images.LoadingError):
            loader(*names)



# Copyright (c) 2021 elParaguayo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import libqtile.config
from libqtile.widget import TextBox
from libqtile.widget.base import _Widget


class TimerWidget(_Widget):
    def cmd_set_timer1(self):
        self.timer1 = self.timeout_add(10, self.cmd_set_timer1)

    def cmd_cancel_timer1(self):
        self.timer1.cancel()

    def cmd_set_timer2(self):
        self.timer2 = self.timeout_add(10, self.cmd_set_timer2)

    def cmd_cancel_timer2(self):
        self.timer2.cancel()

    def cmd_get_active_timers(self):
        active = [x for x in self._futures if x._scheduled]
        return len(active)


def test_multiple_timers(minimal_conf_noscreen, manager_nospawn):
    config = minimal_conf_noscreen
    config.screens = [libqtile.config.Screen(top=libqtile.bar.Bar([TimerWidget(10)], 10))]

    # Start manager and check no active timers
    manager_nospawn.start(config)
    assert manager_nospawn.c.widget["timerwidget"].get_active_timers() == 0

    # Start both timers and confirm both are active
    manager_nospawn.c.widget["timerwidget"].set_timer1()
    manager_nospawn.c.widget["timerwidget"].set_timer2()
    assert manager_nospawn.c.widget["timerwidget"].get_active_timers() == 2

    # Cancel timer1
    manager_nospawn.c.widget["timerwidget"].cancel_timer1()
    assert manager_nospawn.c.widget["timerwidget"].get_active_timers() == 1

    # Cancel timer2
    manager_nospawn.c.widget["timerwidget"].cancel_timer2()
    assert manager_nospawn.c.widget["timerwidget"].get_active_timers() == 0

    # Restart both timers
    manager_nospawn.c.widget["timerwidget"].set_timer1()
    manager_nospawn.c.widget["timerwidget"].set_timer2()
    assert manager_nospawn.c.widget["timerwidget"].get_active_timers() == 2

    # Verify that `finalize()` cancels all timers.
    manager_nospawn.c.widget["timerwidget"].eval("self.finalize()")
    assert manager_nospawn.c.widget["timerwidget"].get_active_timers() == 0


def test_mirrors_same_bar(minimal_conf_noscreen, manager_nospawn):
    """Verify that mirror created when widget reused in same bar."""
    config = minimal_conf_noscreen
    tbox = TextBox("Testing Mirrors")
    config.screens = [libqtile.config.Screen(top=libqtile.bar.Bar([tbox, tbox], 10))]

    manager_nospawn.start(config)
    info = manager_nospawn.c.bar["top"].info()["widgets"]

    # First instance is retained, second is replaced by mirror
    assert len(info) == 2
    assert [w["name"] for w in info] == ["textbox", "mirror"]


def test_mirrors_different_bar(minimal_conf_noscreen, manager_nospawn):
    """Verify that mirror created when widget reused in different bar."""
    config = minimal_conf_noscreen
    tbox = TextBox("Testing Mirrors")
    config.fake_screens = [
        libqtile.config.Screen(top=libqtile.bar.Bar([tbox], 10), x=0, y=0, width=400, height=600),
        libqtile.config.Screen(
            top=libqtile.bar.Bar([tbox], 10), x=400, y=0, width=400, height=600
        ),
    ]

    manager_nospawn.start(config)
    screen0 = manager_nospawn.c.screen[0].bar["top"].info()["widgets"]
    screen1 = manager_nospawn.c.screen[1].bar["top"].info()["widgets"]

    # Original widget should be in the first screen
    assert len(screen0) == 1
    assert [w["name"] for w in screen0] == ["textbox"]

    # Widget is replaced with a mirror on the second screen
    assert len(screen1) == 1
    assert [w["name"] for w in screen1] == ["mirror"]



# Copyright (c) 2008, Aldo Cortesi. All rights reserved.
# Copyright (c) 2011, Florian Mounier
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Set the locale before any widgets or anything are imported, so any widget
# whose defaults depend on a reasonable locale sees something reasonable.
import locale
from os import getenv, makedirs, path
from sys import exit

import libqtile.backend
from libqtile import confreader
from libqtile.log_utils import logger


def rename_process():
    """
    Try to rename the qtile process if py-setproctitle is installed:

    http://code.google.com/p/py-setproctitle/

    Will fail silently if it's not installed. Setting the title lets you do
    stuff like "killall qtile".
    """
    try:
        import setproctitle

        setproctitle.setproctitle("qtile")
    except ImportError:
        pass


def make_qtile(options):
    kore = libqtile.backend.get_core(options.backend)

    if not path.isfile(options.configfile):
        try:
            makedirs(path.dirname(options.configfile), exist_ok=True)
            from shutil import copyfile

            default_config_path = path.join(
                path.dirname(__file__), "..", "resources", "default_config.py"
            )
            copyfile(default_config_path, options.configfile)
            logger.info("Copied default_config.py to %s", options.configfile)
        except Exception as e:
            logger.exception(
                "Failed to copy default_config.py to %s: (%s)", options.configfile, e
            )

    config = confreader.Config(options.configfile)

    # XXX: the import is here because we need to call init_log
    # before start importing stuff
    from libqtile.core.manager import Qtile

    return Qtile(
        kore,
        config,
        no_spawn=options.no_spawn,
        state=options.state,
        socket_path=options.socket,
    )


def start(options):
    try:
        locale.setlocale(locale.LC_ALL, locale.getdefaultlocale())
    except locale.Error:
        pass

    rename_process()
    q = make_qtile(options)
    try:
        q.loop()
    except Exception:
        logger.exception("Qtile crashed")
        exit(1)
    logger.info("Exiting...")


def add_subcommand(subparsers, parents):
    parser = subparsers.add_parser("start", parents=parents, help="Start the window manager")
    parser.add_argument(
        "-c",
        "--config",
        action="store",
        default=path.expanduser(
            path.join(getenv("XDG_CONFIG_HOME", "~/.config"), "qtile", "config.py")
        ),
        dest="configfile",
        help="Use the specified configuration file",
    )
    parser.add_argument(
        "-s",
        "--socket",
        action="store",
        default=None,
        dest="socket",
        help="Path of the Qtile IPC socket.",
    )
    parser.add_argument(
        "-n",
        "--no-spawn",
        action="store_true",
        default=False,
        dest="no_spawn",
        help="Avoid spawning apps. (Used for restart)",
    )
    parser.add_argument(
        "--with-state",
        default=None,
        dest="state",
        help="Pickled QtileState object (typically used only internally)",
    )
    parser.add_argument(
        "-b",
        "--backend",
        default="x11",
        dest="backend",
        choices=libqtile.backend.CORES,
        help="Use specified backend.",
    )
    parser.set_defaults(func=start)
